from unittest.mock import Mock, patch

import numpy as np
import pandas as pd

from sdv.data_processing.numerical_formatter import NumericalFormatter


class TestNumericalFormatter:

    def test___init__(self):
        """Test ``__init__`` attributes properly set."""
        # Run
        formatter = NumericalFormatter(
            enforce_rounding=True,
            enforce_min_max_values=True,
            computer_representation='Int8'
        )

        # Assert
        assert formatter.enforce_rounding is True
        assert formatter.enforce_min_max_values is True
        assert formatter.computer_representation == 'Int8'

    @patch('sdv.data_processing.numerical_formatter.LOGGER')
    def test__learn_rounding_digits_more_than_15_decimals(self, log_mock):
        """Test the ``_learn_rounding_digits`` method with more than 15 decimals.

        If the data has more than 15 decimals, return None and use ``LOGGER`` to inform the user.
        """
        # Setup
        data = pd.Series(np.random.random(size=10).round(20), name='col')

        # Run
        output = NumericalFormatter._learn_rounding_digits(data)

        # Assert
        log_msg = (
            "No rounding scheme detected for column 'col'. Synthetic data will not be rounded."
        )
        log_mock.info.assert_called_once_with(log_msg)
        assert output is None

    def test__learn_rounding_digits_less_than_15_decimals(self):
        """Test the ``_learn_rounding_digits`` method with less than 15 decimals.

        If the data has less than 15 decimals, the maximum number of decimals should be returned.

        Input:
            - an array that contains floats with a maximum of 3 decimals and a NaN.

        Output:
            - 3
        """
        # Setup
        data = pd.Series(np.array([10, 0., 0.1, 0.12, 0.123, np.nan]))

        # Run
        output = NumericalFormatter._learn_rounding_digits(data)

        # Assert
        assert output == 3

    def test__learn_rounding_digits_negative_decimals_float(self):
        """Test the ``_learn_rounding_digits`` method with floats multiples of powers of 10.

        If the data has all multiples of 10 the output should be None.

        Input:
            - an array that contains floats that are multiples of 10, 100 and 1000 and a NaN.
        """
        # Setup
        data = pd.Series(np.array([1230., 12300., 123000., np.nan]))

        # Run
        output = NumericalFormatter._learn_rounding_digits(data)

        # Assert
        assert output == 0

    def test__learn_rounding_digits_negative_decimals_integer(self):
        """Test the ``_learn_rounding_digits`` method with integers multiples of powers of 10.

        If the data has all multiples of 10 the output should be None.

        Input:
            - an array that contains integers that are multiples of 10, 100 and 1000 and a NaN.
        """
        # Setup
        data = pd.Series(np.array([1230, 12300, 123000, np.nan]))

        # Run
        output = NumericalFormatter._learn_rounding_digits(data)

        # Assert
        assert output == 0

    def test__learn_rounding_digits_all_nans(self):
        """Test the ``_learn_rounding_digits`` method with data that is all NaNs.

        If the data is all NaNs, expect that the output is 0.

        Input:
            - an array of NaN.
        """
        # Setup
        data = pd.Series(np.array([np.nan, np.nan, np.nan, np.nan]))

        # Run
        output = NumericalFormatter._learn_rounding_digits(data)

        # Assert
        assert output is None

    def test_learn_format(self):
        """Test that ``learn_format`` method.

        Ensure attributes are correct when ``enforce_min_max_values`` and
        ``enforce_rounding`` are False.

        Setup:
            - a NumericalFormatter with ``_validate_values_within_bounds`` mocked.

        Input:
            - a pandas series.

        Side Effect:
            - only ``_dtype`` is set.
        """
        # Setup
        data = pd.Series([1.5, None, 2.5])
        formatter = NumericalFormatter(enforce_min_max_values=False, enforce_rounding=False)
        formatter._validate_values_within_bounds = Mock()

        # Run
        formatter.learn_format(data)

        # Asserts
        assert formatter._dtype == float
        assert formatter._min_value is None
        assert formatter._max_value is None
        assert formatter._rounding_digits is None

    def test_learn_format_rounding_scheme_true(self):
        """Test ``learn_format`` with ``enforce_rounding`` set to ``True``.

        If ``enforce_rounding`` is set to ``True``, the ``learn_format`` method
        should set its ``_rounding_digits`` instance variable to what is learned
        in the data.

        Input:
            - a Series with floats up to 4 decimals and a None value

        Side Effect:
            - ``_rounding_digits`` is set to 4
        """
        # Setup
        data = pd.Series([1, 2.1, 3.12, 4.123, 5.1234, 6.123, 7.12, 8.1, 9, None])
        formatter = NumericalFormatter(enforce_rounding=True)

        # Run
        formatter.learn_format(data)

        # Asserts
        assert formatter._rounding_digits == 4

    def test__fit_enforce_rounding_true_max_decimals(self):
        """Test ``learn_format`` with ``enforce_rounding`` set to ``True``.

        If the ``enforce_rounding`` parameter is set to ``True``, ``learn_format`` should
        learn the ``_rounding_digits`` to be the max number of decimal places seen in the data.
        The max amount of decimals that floats can be accurately compared with is 15.
        If the input data has values with more than 14 decimals, we will not be able to
        accurately learn the number of decimal places required, so we do not round.

        Input:
            - Series with a value that has 15 decimals
        """
        # Setup
        data = pd.Series([0.000000000000001])
        formatter = NumericalFormatter(enforce_rounding=True)

        # Run
        formatter.learn_format(data)

        # Asserts
        assert formatter._rounding_digits is None

    def test_learn_format_enforce_rounding_true_inf(self):
        """Test ``learn_format`` with ``enforce_rounding`` set to ``True``.

        If the ``enforce_rounding`` parameter is set to ``True``, and the data
        contains only integers or infinite values, ``learn_format`` should learn
        ``_rounding_digits`` to be None.


        Input:
            - Series with ``np.inf`` as a value

        Side Effect:
            - ``_rounding_digits`` is set to None
        """
        # Setup
        data = pd.Series([15000, 4000, 60000, np.inf])
        formatter = NumericalFormatter(enforce_rounding=True)

        # Run
        formatter.learn_format(data)

        # Asserts
        assert formatter._rounding_digits == 0

    def test_learn_format_enforce_rounding_true_max_zero(self):
        """Test ``learn_format`` with ``enforce_rounding`` set to ``True``.

        If the ``enforce_rounding`` parameter is set to ``True``, and the max
        in the data is 0, ``learn_format`` should learn the ``_rounding_digits`` to be None.

        Input:
            - Series with 0 as max value

        Side Effect:
            - ``_rounding_digits`` is set to None
        """
        # Setup
        data = pd.Series([0, 0, 0])
        formatter = NumericalFormatter(enforce_rounding=True)

        # Run
        formatter.learn_format(data)

        # Asserts
        assert formatter._rounding_digits == 0

    def test_learn_format_enforce_min_max_values_true(self):
        """Test ``_fit`` with ``enforce_min_max_values`` set to ``True``.

        If the ``enforce_min_max_values`` parameter is set to ``True``,
        the ``learn_format`` method should learn the min and max values from the data.

        Input:
            - Series of floats and null values

        Side Effect:
            - ``_min_value`` and ``_max_value`` are learned
        """
        # Setup
        data = pd.Series([-100, -5000, 0, None, 100, 4000])
        formatter = NumericalFormatter(enforce_min_max_values=True)

        # Run
        formatter.learn_format(data)

        # Asserts
        assert formatter._min_value == -5000
        assert formatter._max_value == 4000

    def test_format_data_enforce_rounding_false(self):
        """Test ``format_data`` when ``enforce_rounding`` is ``False``.

        The data should not be rounded at all.

        Input:
            - random Series of floats between 0 and 1

        Output:
            - input array
        """
        # Setup
        data = pd.Series(np.random.random(10))
        formatter = NumericalFormatter(enforce_rounding=False)
        formatter._rounding_digits = None

        # Run
        result = formatter.format_data(data)

        # Assert
        np.testing.assert_array_equal(result, data)

    def test_format_data_rounding_none_dtype_int(self):
        """Test ``format_data`` with ``_dtype`` as ``np.int64`` and no rounding.

        The data should be rounded to 0 decimals and returned as integer values if the ``_dtype``
        is ``np.int64`` even if ``_rounding_digits`` is ``None``.

        Input:
            - Series of multiple float values with decimals

        Output:
            - input array rounded an converted to integers
        """
        # Setup
        data = pd.Series([0., 1.2, 3.45, 6.789])
        formatter = NumericalFormatter()
        formatter._rounding_digits = None
        formatter._dtype = np.int64

        # Run
        result = formatter.format_data(data)

        # Assert
        expected = np.array([0, 1, 3, 7])
        np.testing.assert_array_equal(result, expected)

    def test_format_data_rounding_small_numbers(self):
        """Test ``format_data`` when ``_rounding_digits`` is positive.

        The data should round to the maximum number of decimal places
        set in the ``_rounding_digits`` value.

        Input:
            - Series with decimals

        Output:
            - same array rounded to the provided number of decimal places
        """
        # Setup
        data = pd.Series([1.1111, 2.2222, 3.3333, 4.44444, 5.555555])
        formatter = NumericalFormatter()
        formatter.enforce_rounding = True
        formatter._rounding_digits = 2

        # Run
        result = formatter.format_data(data)

        # Assert
        expected_data = np.array([1.11, 2.22, 3.33, 4.44, 5.56])
        np.testing.assert_array_equal(result, expected_data)

    def test_format_data_rounding_big_numbers_type_int(self):
        """Test ``format_data`` when ``_rounding_digits`` is negative.

        The data should round to the number set in the ``_rounding_digits``
        attribute and remain ints.

        Input:
            - Series with with floats above 100

        Output:
            - same array rounded to the provided number of 0s
            - array should be of type int
        """
        # Setup
        data = pd.Series([2000.0, 120.0, 3100.0, 40100.0])
        formatter = NumericalFormatter()
        formatter._dtype = int
        formatter.enforce_rounding = True
        formatter._rounding_digits = -3

        # Run
        result = formatter.format_data(data)

        # Assert
        expected_data = np.array([2000, 0, 3000, 40000])
        np.testing.assert_array_equal(result, expected_data)
        assert result.dtype == int

    def test_format_data_rounding_negative_type_float(self):
        """Test ``format_data`` when ``_rounding_digits`` is negative.

        The data should round to the number set in the ``_rounding_digits``
        attribute and remain floats.

        Input:
            - Series with with larger numbers

        Output:
            - same array rounded to the provided number of 0s
            - array should be of type float
        """
        # Setup
        data = pd.Series([2000.0, 120.0, 3100.0, 40100.0])
        formatter = NumericalFormatter()
        formatter.enforce_rounding = True
        formatter._rounding_digits = -3

        # Run
        result = formatter.format_data(data)

        # Assert
        expected_data = np.array([2000.0, 0.0, 3000.0, 40000.0])
        np.testing.assert_array_equal(result, expected_data)
        assert result.dtype == float

    def test_format_data_rounding_zero_decimal_places(self):
        """Test ``format_data`` when ``_rounding_digits`` is 0.

        The data should round to the number set in the ``_rounding_digits`` attribute.

        Input:
            - Series with with larger numbers

        Output:
            - same array rounded to the 0s place
        """
        # Setup
        data = pd.Series([2000.554, 120.2, 3101, 4010])
        formatter = NumericalFormatter()
        formatter.enforce_rounding = True
        formatter._rounding_digits = 0

        # Run
        result = formatter.format_data(data)

        # Assert
        expected_data = np.array([2001, 120, 3101, 4010])
        np.testing.assert_array_equal(result, expected_data)

    def test_format_data_enforce_min_max_values(self):
        """Test ``format_data`` with ``enforce_min_max_values`` set to ``True``.

        The ``format_data`` method should clip any values above
        the ``max_value`` and any values below the ``min_value``.

        Input:
            - Series with values above the max and below the min

        Output:
            - array with out of bound values clipped to min and max
        """
        # Setup
        data = pd.Series([-np.inf, -5000, -301, -250, 0, 125, 401, np.inf])
        formatter = NumericalFormatter()
        formatter.enforce_min_max_values = True
        formatter._max_value = 400
        formatter._min_value = -300

        # Run
        result = formatter.format_data(data)

        # Asserts
        np.testing.assert_array_equal(result, np.array([-300, -300, -300, -250, 0, 125, 400, 400]))

    def test_format_data_enforce_computer_representation(self):
        """Test ``format_data`` with ``computer_representation`` set to ``Int8``.

        The ``format_data`` method should clip any values out of bounds.

        Input:
            - Series with values above the max and below the min

        Output:
            - array with out of bound values clipped to min and max
        """
        # Setup
        data = pd.Series([-np.inf, np.nan, -5000, -301, -100, 0, 125, 401, np.inf])
        formatter = NumericalFormatter(computer_representation='Int8')

        # Run
        result = formatter.format_data(data)

        # Asserts
        np.testing.assert_array_equal(
            result, np.array([-128, np.nan, -128, -128, -100, 0, 125, 127, 127])
        )



"""Methods to help with upgrading metadata."""

import warnings

from sdv.constraints import (
    FixedCombinations, Inequality, Negative, OneHotEncoding, Positive, Range, ScalarInequality,
    ScalarRange, Unique)


def _upgrade_columns_and_keys(old_metadata):
    new_metadata = {}
    columns = {}
    fields = old_metadata.get('fields')
    alternate_keys = []
    primary_key = old_metadata.get('primary_key')
    for field, field_meta in fields.items():
        column_meta = {}
        old_type = field_meta['type']
        subtype = field_meta.get('subtype')
        column_meta['sdtype'] = old_type

        if old_type == 'numerical':
            if subtype == 'float':
                column_meta['computer_representation'] = 'Float'
            elif subtype == 'integer':
                column_meta['computer_representation'] = 'Int64'

        elif old_type == 'datetime':
            datetime_format = field_meta.get('format')
            if datetime_format:
                column_meta['datetime_format'] = datetime_format

        elif old_type == 'id':
            column_meta['sdtype'] = 'id'
            if subtype == 'integer':
                regex_format = r'\d{30}'
            else:
                regex_format = field_meta.get('regex', '[A-Za-z]{5}')

            if not field_meta.get('pii'):
                column_meta['regex_format'] = regex_format

            if field != primary_key and field_meta.get('ref') is None:
                alternate_keys.append(field)

        if field_meta.get('pii'):
            column_meta['pii'] = True
            pii_category = field_meta.get('pii_category')
            if isinstance(pii_category, list):
                column_meta['sdtype'] = pii_category[0]
            elif pii_category:
                column_meta['sdtype'] = pii_category

        columns[field] = column_meta

    new_metadata['columns'] = columns
    new_metadata['primary_key'] = primary_key
    if alternate_keys:
        new_metadata['alternate_keys'] = alternate_keys

    return new_metadata


def _upgrade_positive_negative(old_constraint):
    new_constraints = []
    strict = old_constraint.get('strict')
    old_name = old_constraint.get('constraint')
    is_positive = old_name == 'sdv.constraints.tabular.Positive'
    constraint_name = Positive.__name__ if is_positive else Negative.__name__
    columns = old_constraint.get('columns')
    if not isinstance(columns, list):
        columns = [columns]

    for column in columns:
        new_constraint = {
            'constraint_name': constraint_name,
            'column_name': column,
            'strict_boundaries': strict
        }
        new_constraints.append(new_constraint)

    return new_constraints


def _upgrade_unique_combinations(old_constraint):
    new_constraint = {
        'constraint_name': FixedCombinations.__name__,
        'column_names': old_constraint.get('columns')
    }

    return [new_constraint]


def _upgrade_greater_than(old_constraint):
    scalar = old_constraint.get('scalar')
    high = old_constraint.get('high')
    low = old_constraint.get('low')
    high_is_string = isinstance(high, str)
    low_is_string = isinstance(low, str)
    strict = old_constraint.get('strict', False)
    new_constraints = []

    if scalar is None:
        high_has_multiple = isinstance(high, list) and len(high) != 1
        low_has_multiple = isinstance(low, list) and len(low) != 1
        high_column_name = f"'{high}'" if high_is_string else high
        low_column_name = f"'{low}'" if low_is_string else low
        if high_has_multiple or low_has_multiple:
            warnings.warn(
                f"Unable to upgrade the GreaterThan constraint specified for 'high' "
                f"{high_column_name} and 'low' {low_column_name}. Manually add "
                f'{Inequality.__name__} constraints to capture this logic.'
            )
            return []

        new_constraint = {
            'constraint_name': Inequality.__name__,
            'high_column_name': high if high_is_string else high[0],
            'low_column_name': low if low_is_string else low[0],
            'strict_boundaries': strict
        }
        new_constraints.append(new_constraint)

    elif scalar == 'low':
        high = [high] if high_is_string else high
        for column in high:
            new_constraint = {
                'constraint_name': ScalarInequality.__name__,
                'column_name': column,
                'relation': '>' if strict else '>=',
                'value': low
            }
            new_constraints.append(new_constraint)

    else:
        low = [low] if low_is_string else low
        for column in low:
            new_constraint = {
                'constraint_name': ScalarInequality.__name__,
                'column_name': column,
                'relation': '<' if strict else '<=',
                'value': high
            }
            new_constraints.append(new_constraint)

    return new_constraints


def _upgrade_between(old_constraint):
    high_is_scalar = old_constraint.get('high_is_scalar')
    low_is_scalar = old_constraint.get('low_is_scalar')
    high = old_constraint.get('high')
    low = old_constraint.get('low')
    constraint_column = old_constraint.get('constraint_column')
    strict = old_constraint.get('strict', False)
    new_constraints = []
    if high_is_scalar and low_is_scalar:
        new_constraint = {
            'constraint_name': ScalarRange.__name__,
            'column_name': constraint_column,
            'low_value': low,
            'high_value': high,
            'strict_boundaries': strict
        }
        new_constraints.append(new_constraint)

    elif high_is_scalar and not low_is_scalar:
        inequality_constraint = {
            'constraint_name': Inequality.__name__,
            'low_column_name': low,
            'high_column_name': constraint_column,
            'strict_boundaries': strict
        }
        scalar_constraint = {
            'constraint_name': ScalarInequality.__name__,
            'column_name': constraint_column,
            'relation': '<' if strict else '<=',
            'value': high
        }
        new_constraints.append(inequality_constraint)
        new_constraints.append(scalar_constraint)

    elif not high_is_scalar and low_is_scalar:
        inequality_constraint = {
            'constraint_name': Inequality.__name__,
            'low_column_name': constraint_column,
            'high_column_name': high,
            'strict_boundaries': strict
        }
        scalar_constraint = {
            'constraint_name': ScalarInequality.__name__,
            'column_name': constraint_column,
            'relation': '>' if strict else '>=',
            'value': low
        }
        new_constraints.append(inequality_constraint)
        new_constraints.append(scalar_constraint)

    else:
        new_constraint = {
            'constraint_name': Range.__name__,
            'low_column_name': low,
            'middle_column_name': constraint_column,
            'high_column_name': high,
            'strict_boundaries': strict
        }
        new_constraints.append(new_constraint)

    return new_constraints


def _upgrade_one_hot_encoding(old_constraint):
    new_constraint = {
        'constraint_name': OneHotEncoding.__name__,
        'column_names': old_constraint.get('columns')
    }
    return [new_constraint]


def _upgrade_unique(old_constraint):
    new_constraint = {
        'constraint_name': Unique.__name__,
        'column_names': old_constraint.get('columns')
    }
    return [new_constraint]


def _upgrade_constraint(old_constraint):
    new_constraints = []
    constraint_name = old_constraint.get('constraint')
    if constraint_name in ('sdv.constraints.tabular.Positive', 'sdv.constraints.tabular.Negative'):
        new_constraints = _upgrade_positive_negative(old_constraint)

    elif constraint_name == 'sdv.constraints.tabular.UniqueCombinations':
        new_constraints = _upgrade_unique_combinations(old_constraint)

    elif constraint_name == 'sdv.constraints.tabular.GreaterThan':
        new_constraints = _upgrade_greater_than(old_constraint)

    elif constraint_name == 'sdv.constraints.tabular.Between':
        new_constraints = _upgrade_between(old_constraint)

    elif constraint_name == 'sdv.constraints.tabular.Rounding':
        warnings.warn(
            "The SDV no longer supports the 'Rounding' constraint. Most models automatically "
            'round to the same number of decimal digits as the original data.'
        )

    elif constraint_name == 'sdv.constraints.tabular.OneHotEncoding':
        new_constraints = _upgrade_one_hot_encoding(old_constraint)

    elif constraint_name == 'sdv.constraints.tabular.Unique':
        new_constraints = _upgrade_unique(old_constraint)

    elif constraint_name == 'sdv.constraints.tabular.ColumnFormula':
        warnings.warn(
            'This method does not upgrade ColumnFormula constraints. Please convert your '
            'logic using the new CustomConstraint API.'
        )

    elif constraint_name == 'sdv.constraints.tabular.CustomConstraint':
        warnings.warn(
            'This method does not upgrade CustomConstraint objects. Please convert '
            'your logic using the new CustomConstraint API.'
        )

    else:
        warnings.warn(
            f'Unable to upgrade the {constraint_name} constraint. Please add in the constraint '
            'using the new Constraints API.'
        )

    return new_constraints


def _upgrade_constraints(old_metadata):
    old_constraints = old_metadata.get('constraints')
    if not old_constraints:
        return None

    needs_upgrading = 'constraint' in old_constraints[0].keys()
    if not needs_upgrading:
        return old_constraints

    new_constraints = []
    for constraint in old_constraints:
        new_constraint = _upgrade_constraint(constraint)
        new_constraints += new_constraint

    return new_constraints


def convert_metadata(old_metadata):
    """Convert old metadata to the new metadata format.

    Args:
        old_metadata (dict):
            A dict version of the old metadata.

    Returns:
        An equivalent dict in the new metadata format.
    """
    new_metadata = _upgrade_columns_and_keys(old_metadata)
    return new_metadata



import numpy as np
import pandas as pd

from sdv.lite import SingleTablePreset
from sdv.metadata import SingleTableMetadata


def test_sample():
    """Test sampling for the ``SingleTablePreset``."""
    # Setup
    data = pd.DataFrame({'a': [1, 2, 3, np.nan]})

    # Run
    metadata = SingleTableMetadata()
    metadata.detect_from_dataframe(data)
    preset = SingleTablePreset(metadata, name='FAST_ML')
    preset.fit(data)
    samples = preset.sample(
        num_rows=10,
        max_tries_per_batch=20,
        batch_size=5
    )

    # Assert
    assert samples['a'].all() in [1, 2, 3, np.nan]
    assert len(samples) == 10


def test_sample_with_constraints():
    """Test sampling for the ``SingleTablePreset``."""
    # Setup
    data = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6]
    })

    # Run
    metadata = SingleTableMetadata()
    metadata.detect_from_dataframe(data)
    preset = SingleTablePreset(metadata, name='FAST_ML')
    constraints = [
        {
            'constraint_class': 'Inequality',
            'constraint_parameters': {
                'low_column_name': 'a',
                'high_column_name': 'b'
            }
        }
    ]
    preset.add_constraints(constraints)
    preset.fit(data)
    samples = preset.sample(
        num_rows=10,
        max_tries_per_batch=20,
        batch_size=5
    )

    # Assert
    assert len(samples) == 10
    assert all(samples['a'] < samples['b'])



"""PAR Synthesizer class."""

import inspect
import logging
import uuid
import warnings

import numpy as np
import pandas as pd
import tqdm
from deepecho import PARModel
from deepecho.sequences import assemble_sequences

from sdv.errors import SamplingError, SynthesizerInputError
from sdv.metadata.single_table import SingleTableMetadata
from sdv.single_table import GaussianCopulaSynthesizer
from sdv.single_table.base import BaseSynthesizer
from sdv.utils import cast_to_iterable

LOGGER = logging.getLogger(__name__)


class PARSynthesizer(BaseSynthesizer):
    """Synthesizer for sequential data.

    This synthesizer uses the ``deepecho.models.par.PARModel`` class as the core model.
    Additionally, it uses a separate synthesizer to model and sample the context columns
    to be passed into PAR.

    Args:
        metadata (sdv.metadata.SingleTableMetadata):
            Single table metadata representing the data that this synthesizer will be used for.
        enforce_min_max_values (bool):
            Specify whether or not to clip the data returned by ``reverse_transform`` of
            the numerical transformer, ``FloatFormatter``, to the min and max values seen
            during ``fit``. Defaults to ``True``.
        enforce_rounding (bool):
            Define rounding scheme for ``numerical`` columns. If ``True``, the data returned
            by ``reverse_transform`` will be rounded as in the original data. Defaults to ``True``.
        context_columns (list[str]):
            A list of strings, representing the columns that do not vary in a sequence.
        segment_size (int):
            If specified, cut each training sequence in several segments of
            the indicated size. The size can be passed as an integer
            value, which will interpreted as the number of data points to
            put on each segment.
        epochs (int):
            The number of epochs to train for. Defaults to 128.
        sample_size (int):
            The number of times to sample (before choosing and
            returning the sample which maximizes the likelihood).
            Defaults to 1.
        cuda (bool):
            Whether to attempt to use cuda for GPU computation.
            If this is False or CUDA is not available, CPU will be used.
            Defaults to ``True``.
        verbose (bool):
            Whether to print progress to console or not.
    """

    _model_sdtype_transformers = {
        'categorical': None,
        'numerical': None,
        'boolean': None
    }

    def _get_context_metadata(self):
        context_columns_dict = {}
        context_columns = self.context_columns.copy() if self.context_columns else []
        if self._sequence_key:
            context_columns += self._sequence_key

        for column in context_columns:
            context_columns_dict[column] = self.metadata.columns[column]

        context_metadata_dict = {'columns': context_columns_dict}
        return SingleTableMetadata.load_from_dict(context_metadata_dict)

    def __init__(self, metadata, enforce_min_max_values=True, enforce_rounding=False,
                 context_columns=None, segment_size=None, epochs=128, sample_size=1, cuda=True,
                 verbose=False):
        super().__init__(
            metadata=metadata,
            enforce_min_max_values=enforce_min_max_values,
            enforce_rounding=enforce_rounding,
        )

        sequence_key = self.metadata.sequence_key
        self._sequence_key = list(cast_to_iterable(sequence_key)) if sequence_key else None
        if context_columns and not self._sequence_key:
            raise SynthesizerInputError(
                "No 'sequence_keys' are specified in the metadata. The PARSynthesizer cannot "
                "model 'context_columns' in this case."
            )

        self._sequence_index = self.metadata.sequence_index
        self.context_columns = context_columns or []
        self.segment_size = segment_size
        self._model_kwargs = {
            'epochs': epochs,
            'sample_size': sample_size,
            'cuda': cuda,
            'verbose': verbose,
        }
        context_metadata = self._get_context_metadata()
        self._context_synthesizer = GaussianCopulaSynthesizer(
            metadata=context_metadata,
            enforce_min_max_values=enforce_min_max_values,
            enforce_rounding=enforce_rounding
        )

    def get_parameters(self):
        """Return the parameters used to instantiate the synthesizer."""
        parameters = inspect.signature(self.__init__).parameters
        instantiated_parameters = {}
        for parameter_name in parameters:
            if parameter_name != 'metadata':
                instantiated_parameters[parameter_name] = self.__dict__.get(parameter_name)

        for parameter_name, value in self._model_kwargs.items():
            instantiated_parameters[parameter_name] = value

        return instantiated_parameters

    def add_constraints(self, constraints):
        """Warn the user that constraints can't be added to the ``PARSynthesizer``."""
        warnings.warn(
            'The PARSynthesizer does not yet support constraints. This model will ignore any '
            'constraints in the metadata.'
        )
        self._data_processor._constraints = []
        self._data_processor._constraints_list = []

    def _validate_context_columns(self, data):
        errors = []
        if self.context_columns:
            errors = []
            for sequence_key_value, data_values in data.groupby(self._sequence_key):
                for context_column in self.context_columns:
                    if len(data_values[context_column].unique()) > 1:
                        errors.append((
                            f"Context column '{context_column}' is changing inside sequence "
                            f'({self._sequence_key}={sequence_key_value}).'
                        ))

        return errors

    def _validate(self, data):
        return self._validate_context_columns(data)

    def _preprocess(self, data):
        """Transform the raw data to numerical space.

        For PAR, none of the sequence keys are transformed.

        Args:
            data (pandas.DataFrame):
                The raw data to be transformed.

        Returns:
            pandas.DataFrame:
                The preprocessed data.
        """
        sequence_key_transformers = {sequence_key: None for sequence_key in self._sequence_key}
        if not self._data_processor._prepared_for_fitting:
            self.auto_assign_transformers(data)

        self.update_transformers(sequence_key_transformers)
        return super()._preprocess(data)

    def update_transformers(self, column_name_to_transformer):
        """Update any of the transformers assigned to each of the column names.

        Args:
            column_name_to_transformer (dict):
                Dict mapping column names to transformers to be used for that column.

        Raises:
            ValueError:
                Raise when the transformer of a context column is passed.
        """
        if set(column_name_to_transformer).intersection(set(self.context_columns)):
            raise SynthesizerInputError(
                'Transformers for context columns are not allowed to be updated.')

        super().update_transformers(column_name_to_transformer)

    def _fit_context_model(self, transformed):
        LOGGER.debug(f'Fitting context synthesizer {self._context_synthesizer.__class__.__name__}')
        if self.context_columns:
            context = transformed[self._sequence_key + self.context_columns]
        else:
            context = transformed[self._sequence_key].copy()
            # Add constant column to allow modeling
            constant_column = str(uuid.uuid4())
            context[constant_column] = 0
            self._context_synthesizer.metadata.add_column(constant_column, sdtype='numerical')

        context = context.groupby(self._sequence_key).first().reset_index()
        self._context_synthesizer.fit(context)

    def _transform_sequence_index(self, sequences):
        sequence_index_idx = self._data_columns.index(self._sequence_index)
        for sequence in sequences:
            data = sequence['data']
            sequence_index = data[sequence_index_idx]
            diffs = np.diff(sequence_index).tolist()
            data[sequence_index_idx] = diffs[0:1] + diffs
            data.append(sequence_index[0:1] * len(sequence_index))

    def _fit_sequence_columns(self, timeseries_data):
        self._model = PARModel(**self._model_kwargs)

        self._output_columns = list(timeseries_data.columns)
        self._data_columns = [
            column
            for column in timeseries_data.columns
            if column not in self._sequence_key + self.context_columns
        ]

        sequences = assemble_sequences(
            timeseries_data,
            self._sequence_key,
            self.context_columns,
            self.segment_size,
            self._sequence_index,
            drop_sequence_index=False
        )
        data_types = []
        context_types = []
        for field in self._output_columns:
            dtype = timeseries_data[field].dtype
            kind = dtype.kind
            if kind in ('i', 'f'):
                data_type = 'continuous'
            elif kind in ('O', 'b'):
                data_type = 'categorical'
            else:
                raise ValueError(f'Unsupported dtype {dtype}')

            if field in self._data_columns:
                data_types.append(data_type)
            elif field in self.context_columns:
                context_types.append(data_type)

        if self._sequence_index:
            self._transform_sequence_index(sequences)
            data_types.append('continuous')

        # Validate and fit
        self._model.fit_sequences(sequences, context_types, data_types)

    def _fit(self, processed_data):
        """Fit this model to the data.

        Args:
            processed_data (pandas.DataFrame):
                pandas.DataFrame containing both the sequences,
                the entity columns and the context columns.
        """
        if self._sequence_key:
            self._fit_context_model(processed_data)

        LOGGER.debug(f'Fitting {self.__class__.__name__} model to table')
        self._fit_sequence_columns(processed_data)

    def _sample_from_par(self, context, sequence_length=None):
        """Sample new sequences.

        Args:
            context (pandas.DataFrame):
                Context values to use when generating the sequences.
            sequence_length (int):
                Length of each sequence to sample. If not
                given, the sequence length will be sampled from
                the model.

        Returns:
            pandas.DataFrame:
                Table containing the sampled sequences in the same
                format as that he training data had.
        """
        # Set the sequence_key as index to properly iterate over them
        if self._sequence_key:
            context = context.set_index(self._sequence_key)
            # reorder context columns
            context = context[self.context_columns]

        should_disable = not self._model_kwargs['verbose']
        iterator = tqdm.tqdm(context.iterrows(), disable=should_disable, total=len(context))

        output = []
        for sequence_key_values, context_values in iterator:
            context_values = context_values.tolist()
            sequence = self._model.sample_sequence(context_values, sequence_length)
            if self._sequence_index:
                sequence_index_idx = self._data_columns.index(self._sequence_index)
                diffs = sequence[sequence_index_idx]
                start = sequence.pop(-1)
                sequence[sequence_index_idx] = np.cumsum(diffs) - diffs[0] + start

            # Reformat as a DataFrame
            sequence_df = pd.DataFrame(
                dict(zip(self._data_columns, sequence)),
                columns=self._data_columns
            )
            sequence_df[self._sequence_key] = sequence_key_values
            for column, value in zip(self.context_columns, context_values):
                sequence_df[column] = value

            output.append(sequence_df)

        output = pd.concat(output)
        output = output[self._output_columns].reset_index(drop=True)

        return output

    def _sample(self, context_columns, sequence_length=None):
        sampled = self._sample_from_par(context_columns, sequence_length)
        return self._data_processor.reverse_transform(sampled)

    def sample(self, num_sequences, sequence_length=None):
        """Sample new sequences.

        Args:
            num_sequences (int):
                Number of sequences to sample.
            sequence_length (int):
                If passed, sample sequences of this length. If ``None``, the sequence length will
                be sampled from the model.

        Returns:
            pandas.DataFrame:
                Table containing the sampled sequences in the same format as the fitted data.
        """
        if self._sequence_key:
            context_columns = self._context_synthesizer._sample_with_progress_bar(
                num_sequences,
                output_file_path='disable',
                show_progress_bar=False
            )

        else:
            context_columns = pd.DataFrame(index=range(num_sequences or 1))

        for column in self._sequence_key or []:
            if column not in context_columns:
                context_columns[column] = range(len(context_columns))

        return self._sample(context_columns, sequence_length)

    def sample_sequential_columns(self, context_columns, sequence_length=None):
        """Sample the sequential columns based ont he provided context columns.

        Args:
            context_columns (pandas.DataFrame):
                Context values to use when generating the sequences.
            sequence_length (int):
                If passed, sample sequences of this length. If ``None``, the sequence length will
                be sampled from the model.

        Returns:
            pandas.DataFrame:
                Table containing the sampled sequences based on the provided context columns.
        """
        if not self.context_columns:
            raise SamplingError(
                "This synthesizer does not have any context columns. Please use 'sample()' "
                'to sample new sequences.'
            )

        return self._sample(context_columns, sequence_length)



"""Integration tests for Single Table Metadata."""

import json
import re

import pytest

from sdv.metadata import SingleTableMetadata
from sdv.metadata.errors import InvalidMetadataError


def test_single_table_metadata():
    """Test ``SingleTableMetadata``."""

    # Create an instance
    instance = SingleTableMetadata()

    # To dict
    result = instance.to_dict()

    # Assert
    assert result == {
        'METADATA_SPEC_VERSION': 'SINGLE_TABLE_V1'
    }
    assert instance.columns == {}
    assert instance._version == 'SINGLE_TABLE_V1'
    assert instance.primary_key is None
    assert instance.sequence_key is None
    assert instance.alternate_keys == []
    assert instance.sequence_index is None


def test_validate():
    """Test ``SingleTableMetadata.validate``.

    Ensure the method doesn't crash for a valid metadata.
    """
    # Setup
    instance = SingleTableMetadata()
    instance.add_column('col1', sdtype='id')
    instance.add_column('col2', sdtype='id')
    instance.add_column('col3', sdtype='numerical')
    instance.set_primary_key('col1')
    instance.add_alternate_keys([('col1', 'col2')])
    instance.set_sequence_index('col3')
    instance.set_sequence_key('col2')

    # Run
    instance.validate()


def test_validate_errors():
    """Test ``SingleTableMetadata.validate`` raises the correct errors."""
    # Setup
    instance = SingleTableMetadata()
    instance.columns = {
        'col1': {'sdtype': 'id'},
        'col2': {'sdtype': 'numerical'},
        'col4': {'sdtype': 'categorical', 'invalid1': 'value'},
        'col5': {'sdtype': 'categorical', 'order': ''},
        'col6': {'sdtype': 'categorical', 'order_by': ''},
        'col7': {'sdtype': 'categorical', 'order': '', 'order_by': ''},
        'col8': {'sdtype': 'numerical', 'computer_representation': 'value'},
        'col9': {'sdtype': 'datetime', 'datetime_format': '%1-%Y-%m-%d-%'},
        'col10': {'sdtype': 'id', 'regex_format': '[A-{6}'},
    }
    instance.primary_key = 10
    instance.alternate_keys = 'col1'
    instance.sequence_key = ('col3', 'col1')
    instance.sequence_index = 'col3'

    err_msg = re.escape(
        'The following errors were found in the metadata:'
        "\n\n'primary_key' must be a string or tuple of strings."
        "\nUnknown sequence key values {'col3'}. Keys should be columns that exist in the table."
        "\nUnknown sequence index value {'col3'}. Keys should be columns that exist in the table."
        "\n'sequence_index' and 'sequence_key' have the same value {'col3'}."
        ' These columns must be different.'
        "\n'alternate_keys' must be a list of strings or a list of tuples of strings."
        "\nInvalid values '(invalid1)' for categorical column 'col4'."
        "\nInvalid order value provided for categorical column 'col5'."
        " The 'order' must be a list with 1 or more elements."
        "\nUnknown ordering method '' provided for categorical column 'col6'."
        " Ordering method must be 'numerical_value' or 'alphabetical'."
        "\nCategorical column 'col7' has both an 'order' and 'order_by' attribute."
        ' Only 1 is allowed.'
        "\nInvalid value for 'computer_representation' 'value' for column 'col8'."
        "\nInvalid datetime format string '%1-%Y-%m-%d-%' for datetime column 'col9'."
        "\nInvalid regex format string '[A-{6}' for id column 'col10'."
    )
    # Run / Assert
    with pytest.raises(InvalidMetadataError, match=err_msg):
        instance.validate()


def test_upgrade_metadata(tmp_path):
    """Test the ``upgrade_metadata`` method."""
    # Setup
    old_metadata = {
        'fields': {
            'start_date': {
                'type': 'datetime',
                'format': '%Y-%m-%d'
            },
            'end_date': {
                'type': 'datetime',
                'format': '%Y-%m-%d'
            },
            'salary': {
                'type': 'numerical',
                'subtype': 'integer'
            },
            'duration': {
                'type': 'categorical'
            },
            'student_id': {
                'type': 'id',
                'subtype': 'integer'
            },
            'high_perc': {
                'type': 'numerical',
                'subtype': 'float'
            },
            'placed': {
                'type': 'boolean'
            },
            'ssn': {
                'type': 'id',
                'subtype': 'integer'
            },
            'drivers_license': {
                'type': 'id',
                'subtype': 'string',
                'regex': 'regex'
            }
        },
        'primary_key': 'student_id'
    }
    filepath = tmp_path / 'old.json'
    old_metadata_file = open(filepath, 'w')
    json.dump(old_metadata, old_metadata_file)
    old_metadata_file.close()

    # Run
    new_metadata = SingleTableMetadata.upgrade_metadata(filepath=filepath).to_dict()

    # Assert
    expected_metadata = {
        'columns': {
            'start_date': {
                'sdtype': 'datetime',
                'datetime_format': '%Y-%m-%d'
            },
            'end_date': {
                'sdtype': 'datetime',
                'datetime_format': '%Y-%m-%d'
            },
            'salary': {
                'sdtype': 'numerical',
                'computer_representation': 'Int64'
            },
            'duration': {
                'sdtype': 'categorical'
            },
            'student_id': {
                'sdtype': 'id',
                'regex_format': r'\d{30}'
            },
            'high_perc': {
                'sdtype': 'numerical',
                'computer_representation': 'Float'
            },
            'placed': {
                'sdtype': 'boolean'
            },
            'ssn': {
                'sdtype': 'id',
                'regex_format': r'\d{30}'
            },
            'drivers_license': {
                'sdtype': 'id',
                'regex_format': 'regex'
            }
        },
        'primary_key': 'student_id',
        'alternate_keys': ['ssn', 'drivers_license'],
        'METADATA_SPEC_VERSION': 'SINGLE_TABLE_V1'
    }
    assert new_metadata == expected_metadata



from unittest.mock import call, patch

from sdv.metadata.metadata_upgrader import _upgrade_constraints, convert_metadata


def test__upgrade_constraints_no_constraints():
    """Test the ``_upgrade_constraints`` method with no constraints.

    Input:
        - Metadata dict with no constraints.

    Output:
        - None.
    """
    # Setup
    old_metadata = {'fields': {}}

    # Run
    new_constraints = _upgrade_constraints(old_metadata)

    # Assert
    assert new_constraints is None


def test__upgrade_constraints_no_upgrade_needed():
    """Test the ``_upgrade_constraints`` method when new constraints are already used.

    Input:
        - Old metadata with new constraints.

    Output:
        - Same as input.
    """
    # Setup
    old_constraints = [
        {
            'constraint_name': 'OneHotEncoding',
            'column_names': ['a', 'b']
        },
        {
            'constraint_name': 'Unique',
            'column_names': ['c', 'd']
        },
    ]
    old_metadata = {'constraints': old_constraints}

    # Run
    new_constraints = _upgrade_constraints(old_metadata)

    # Assert
    assert old_constraints == new_constraints


def test__upgrade_constraints_greater_than():
    """Test the ``_upgrade_constraints`` method with ``GreaterThan`` constraints.

    Input:
        - Old metadata dict with the following constraints:
        - ``GreaterThan`` with ``scalar`` set to None.
        - ``GreaterThan`` with ``scalar`` set to 'high'.
        - ``GreaterThan`` with ``scalar`` set to 'high' and multiple columns.
        - ``GreaterThan`` with ``scalar`` set to 'low'.
        - ``GreaterThan`` with ``scalar`` set to 'low' and multiple columns.

    Ouput:
        - Metadata with the following constraints:
        - ``Inequality``.
        - ``ScalarInequality`` constraints with the value as low and relation as '>'.
        - ``ScalarInequality`` constraints with the value as high and relation as '>'.
    """
    # Setup
    old_constraints = [
        {
            'constraint': 'sdv.constraints.tabular.GreaterThan',
            'scalar': None,
            'high': 'a',
            'low': 'b',
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.GreaterThan',
            'scalar': None,
            'high': ['c'],
            'low': 'd',
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.GreaterThan',
            'scalar': None,
            'high': 'e',
            'low': ['f'],
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.GreaterThan',
            'scalar': 'low',
            'high': 'a',
            'low': 10
        },
        {
            'constraint': 'sdv.constraints.tabular.GreaterThan',
            'scalar': 'low',
            'high': ['b', 'c'],
            'low': 5
        },
        {
            'constraint': 'sdv.constraints.tabular.GreaterThan',
            'scalar': 'high',
            'high': 10,
            'low': 'f'
        },
        {
            'constraint': 'sdv.constraints.tabular.GreaterThan',
            'scalar': 'high',
            'high': 5,
            'low': ['c', 'd']
        }
    ]
    old_metadata = {'constraints': old_constraints}

    # Run
    new_constraints = _upgrade_constraints(old_metadata)

    # Assert
    expected_constraints = [
        {
            'constraint_name': 'Inequality',
            'high_column_name': 'a',
            'low_column_name': 'b',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Inequality',
            'high_column_name': 'c',
            'low_column_name': 'd',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Inequality',
            'high_column_name': 'e',
            'low_column_name': 'f',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'ScalarInequality',
            'column_name': 'a',
            'relation': '>=',
            'value': 10
        },
        {
            'constraint_name': 'ScalarInequality',
            'column_name': 'b',
            'relation': '>=',
            'value': 5
        },
        {
            'constraint_name': 'ScalarInequality',
            'column_name': 'c',
            'relation': '>=',
            'value': 5
        },
        {
            'constraint_name': 'ScalarInequality',
            'column_name': 'f',
            'relation': '<=',
            'value': 10
        },
        {
            'constraint_name': 'ScalarInequality',
            'column_name': 'c',
            'relation': '<=',
            'value': 5
        },
        {
            'constraint_name': 'ScalarInequality',
            'column_name': 'd',
            'relation': '<=',
            'value': 5
        }
    ]

    assert len(expected_constraints) == len(new_constraints)
    for constraint in expected_constraints:
        assert constraint in new_constraints


@patch('sdv.metadata.metadata_upgrader.warnings')
def test__upgrade_constraints_greater_than_error(warnings_mock):
    """Test the ``_upgrade_constraints`` method with ``GreaterThan`` constraints.

    Input:
        - Old metadata dict with the following constraints:
        - ``GreaterThan`` with one of high or low being a list of multiple strings.

    Ouput:
        - Empty list.

    Side effect:
        - Warnings should be called.
    """
    # Setup
    old_constraints = [
        {
            'constraint': 'sdv.constraints.tabular.GreaterThan',
            'scalar': None,
            'high': ['a', 'b'],
            'low': 'c',
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.GreaterThan',
            'scalar': None,
            'high': 'a',
            'low': ['b', 'c'],
            'strict': True
        }
    ]
    old_metadata = {'constraints': old_constraints}

    # Run
    new_constraints = _upgrade_constraints(old_metadata)

    # Assert
    assert new_constraints == []
    warnings_mock.warn.assert_has_calls([
        call(
            "Unable to upgrade the GreaterThan constraint specified for 'high' ['a', 'b'] "
            "and 'low' 'c'. Manually add Inequality constraints to capture this logic."
        ),
        call(
            "Unable to upgrade the GreaterThan constraint specified for 'high' 'a' "
            "and 'low' ['b', 'c']. Manually add Inequality constraints to capture this logic."
        )
    ])


def test__upgrade_constraints_between():
    """Test the ``_upgrade_constraints`` method with ``Between`` constraints.

    Input:
        - Old metadata dict with the following constraints:
        - ``Between`` constraint with 'high_is_scalar' and 'low_is_scalar' set to True.
        - ``Between`` constraint with 'high_is_scalar' and 'low_is_scalar' set to False.
        - ``Between`` constraint with 'high_is_scalar' as True and 'low_is_scalar' as False.
        - ``Between`` constraint with 'high_is_scalar' as False and 'low_is_scalar' as True.

    Output:
        - Metadata with the following constraints:
        - ``ScalarRange``.
        - ``Range``.
        - ``Inequality`` paired with a ``ScalarInequality`` with relation set to '<='.
        - ``Inequality`` paired with a ``ScalarInequality`` with relation set to '>='.
    """
    # Setup
    old_constraints = [
        {
            'constraint': 'sdv.constraints.tabular.Between',
            'constraint_column': 'z',
            'high_is_scalar': True,
            'low_is_scalar': True,
            'low': 5,
            'high': 10,
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.Between',
            'constraint_column': 'z',
            'high_is_scalar': False,
            'low_is_scalar': False,
            'low': 'a',
            'high': 'b',
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.Between',
            'constraint_column': 'z',
            'high_is_scalar': True,
            'low_is_scalar': False,
            'low': 'a',
            'high': 10
        },
        {
            'constraint': 'sdv.constraints.tabular.Between',
            'constraint_column': 'z',
            'high_is_scalar': False,
            'low_is_scalar': True,
            'low': 5,
            'high': 'b'
        }
    ]
    old_metadata = {'constraints': old_constraints}

    # Run
    new_constraints = _upgrade_constraints(old_metadata)

    # Assert
    expected_constraints = [
        {
            'constraint_name': 'ScalarRange',
            'column_name': 'z',
            'low_value': 5,
            'high_value': 10,
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Range',
            'middle_column_name': 'z',
            'low_column_name': 'a',
            'high_column_name': 'b',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Inequality',
            'low_column_name': 'a',
            'high_column_name': 'z',
            'strict_boundaries': False
        },
        {
            'constraint_name': 'ScalarInequality',
            'column_name': 'z',
            'relation': '<=',
            'value': 10,
        },
        {
            'constraint_name': 'Inequality',
            'low_column_name': 'z',
            'high_column_name': 'b',
            'strict_boundaries': False
        },
        {
            'constraint_name': 'ScalarInequality',
            'column_name': 'z',
            'relation': '>=',
            'value': 5,
        }
    ]
    assert len(expected_constraints) == len(new_constraints)
    for constraint in expected_constraints:
        assert constraint in new_constraints


def test__upgrade_constraints_positive_and_negative():
    """Test the ``_upgrade_constraints`` method with ``Positive`` and ``Negative`` constraints.

    Input:
        - Old metadata dict with the following constraints:
        - ``Positive`` constraint with one column and 'strict' as True.
        - ``Positive`` constraint with one column and 'strict' as False.
        - ``Positive`` constraint with multiple columns and 'strict' as True.
        - ``Positive`` constraint with multiple columns and 'strict' as False.
        - ``Negative`` constraint with one column and 'strict' as True.
        - ``Negative`` constraint with one column and 'strict' as False.
        - ``Negative`` constraint with multiple columns and 'strict' as True.
        - ``Negative`` constraint with multiple columns and 'strict' as False.

    Output:
        - Metadata with the following constraints:
        - ``Positive`` constraint for every ``Positive`` that had strict as True.
        - ``ScalarInequality`` for every ``Positive`` that had strict as False.
        - ``Negative`` constraint for every ``Negative`` that had strict as True.
        - ``ScalarInequality`` for every ``Negative`` that had strict as False.
    """
    # Setup
    old_constraints = [
        {
            'constraint': 'sdv.constraints.tabular.Positive',
            'columns': 'a',
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.Positive',
            'columns': ['b', 'c'],
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.Positive',
            'columns': 'd',
            'strict': False
        },
        {
            'constraint': 'sdv.constraints.tabular.Positive',
            'columns': ['e', 'f'],
            'strict': False
        },
        {
            'constraint': 'sdv.constraints.tabular.Negative',
            'columns': 'a',
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.Negative',
            'columns': ['b', 'c'],
            'strict': True
        },
        {
            'constraint': 'sdv.constraints.tabular.Negative',
            'columns': 'd',
            'strict': False
        },
        {
            'constraint': 'sdv.constraints.tabular.Negative',
            'columns': ['e', 'f'],
            'strict': False
        },
    ]
    old_metadata = {'constraints': old_constraints}

    # Run
    new_constraints = _upgrade_constraints(old_metadata)

    # Assert
    expected_constraints = [
        {
            'constraint_name': 'Positive',
            'column_name': 'a',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Positive',
            'column_name': 'b',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Positive',
            'column_name': 'c',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Positive',
            'column_name': 'd',
            'strict_boundaries': False
        },
        {
            'constraint_name': 'Positive',
            'column_name': 'e',
            'strict_boundaries': False
        },
        {
            'constraint_name': 'Positive',
            'column_name': 'f',
            'strict_boundaries': False
        },
        {
            'constraint_name': 'Negative',
            'column_name': 'a',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Negative',
            'column_name': 'b',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Negative',
            'column_name': 'c',
            'strict_boundaries': True
        },
        {
            'constraint_name': 'Negative',
            'column_name': 'd',
            'strict_boundaries': False
        },
        {
            'constraint_name': 'Negative',
            'column_name': 'e',
            'strict_boundaries': False
        },
        {
            'constraint_name': 'Negative',
            'column_name': 'f',
            'strict_boundaries': False
        }
    ]
    assert len(expected_constraints) == len(new_constraints)
    for constraint in expected_constraints:
        assert constraint in new_constraints


def test__upgrade_constraints_simple_constraints():
    """Test the ``_upgrade_constraints`` method with constraints that are easy to convert.

    Input:
        - Old metadata with the following constraints:
        - A ``UniqueCombinations`` constraint.
        - A ``OneHotEncoding`` constraint.
        - A ``Unique`` constraint.

    Ouput:
        - A ``FixedCombinations`` constraint.
        - A ``OneHotEncoding`` constraint.
        - A ``Unique`` constraint.
    """
    # Setup
    old_constraints = [
        {
            'constraint': 'sdv.constraints.tabular.UniqueCombinations',
            'columns': ['a', 'b']
        },
        {
            'constraint': 'sdv.constraints.tabular.OneHotEncoding',
            'columns': ['c', 'd']
        },
        {
            'constraint': 'sdv.constraints.tabular.Unique',
            'columns': ['e', 'f']
        },
    ]
    old_metadata = {'constraints': old_constraints}

    # Run
    new_constraints = _upgrade_constraints(old_metadata)

    # Assert
    expected_constraints = [
        {
            'constraint_name': 'FixedCombinations',
            'column_names': ['a', 'b']
        },
        {
            'constraint_name': 'OneHotEncoding',
            'column_names': ['c', 'd']
        },
        {
            'constraint_name': 'Unique',
            'column_names': ['e', 'f']
        },
    ]
    assert len(expected_constraints) == len(new_constraints)
    for constraint in expected_constraints:
        assert constraint in new_constraints


@patch('sdv.metadata.metadata_upgrader.warnings')
def test__upgrade_constraints_constraint_has_no_upgrade(warnings_mock):
    """Test the ``_upgrade_constraints`` method with constraints that can't be upgraded.

    Setup:
        - Patch warnings.

    Input:
        - Old metadata with following constraints:
        - ``Rounding``.
        - ``ColumnFormula``.
        - ``CustomConstraint``.
        - A fake constraint.

    Output:
        - An empty list.

    Side effect:
        - Warnings should be raised.
    """
    # Setup
    old_constraints = [
        {
            'constraint': 'sdv.constraints.tabular.Rounding'
        },
        {
            'constraint': 'sdv.constraints.tabular.ColumnFormula'
        },
        {
            'constraint': 'sdv.constraints.tabular.CustomConstraint'
        },
        {
            'constraint': 'Fake'
        }
    ]
    old_metadata = {'constraints': old_constraints}

    # Run
    new_constraints = _upgrade_constraints(old_metadata)

    # Assert
    assert new_constraints == []
    assert len(warnings_mock.warn.mock_calls) == 4
    warnings_mock.warn.assert_has_calls([
        call(
            "The SDV no longer supports the 'Rounding' constraint. Most models automatically "
            'round to the same number of decimal digits as the original data.'
        ),
        call(
            'This method does not upgrade ColumnFormula constraints. Please convert your '
            'logic using the new CustomConstraint API.'
        ),
        call(
            'This method does not upgrade CustomConstraint objects. Please convert '
            'your logic using the new CustomConstraint API.'
        ),
        call(
            'Unable to upgrade the Fake constraint. Please add in the constraint '
            'using the new Constraints API.'
        )
    ])


def test_convert_metadata():
    """Test the ``convert_metadata`` method.

    The method should take a dictionary of the old metadata format and convert it to the new
    format.

    Input:
        - Dictionary of single table metadata in the old schema.

    Output:
        - Dictionary of the same metadata with the new schema.
    """
    # Setup
    old_metadata = {
        'fields': {
            'start_date': {
                'type': 'datetime',
                'format': '%Y-%m-%d'
            },
            'end_date': {
                'type': 'datetime',
                'format': '%Y-%m-%d'
            },
            'salary': {
                'type': 'numerical',
                'subtype': 'integer'
            },
            'duration': {
                'type': 'categorical'
            },
            'student_id': {
                'type': 'id',
                'subtype': 'integer'
            },
            'high_perc': {
                'type': 'numerical',
                'subtype': 'float'
            },
            'placed': {
                'type': 'boolean'
            },
            'ssn': {
                'type': 'categorical',
                'pii': True,
                'pii_category': 'ssn'
            },
            'credit_card': {
                'type': 'categorical',
                'pii': True,
                'pii_category': ['credit_card_number', 'visa']
            },
            'drivers_license': {
                'type': 'id',
                'subtype': 'string',
                'regex': 'regex'
            }
        },
        'primary_key': 'student_id'
    }

    # Run
    new_metadata = convert_metadata(old_metadata)

    # Assert
    expected_metadata = {
        'columns': {
            'start_date': {
                'sdtype': 'datetime',
                'datetime_format': '%Y-%m-%d'
            },
            'end_date': {
                'sdtype': 'datetime',
                'datetime_format': '%Y-%m-%d'
            },
            'salary': {
                'sdtype': 'numerical',
                'computer_representation': 'Int64'
            },
            'duration': {
                'sdtype': 'categorical'
            },
            'student_id': {
                'sdtype': 'id',
                'regex_format': r'\d{30}'
            },
            'high_perc': {
                'sdtype': 'numerical',
                'computer_representation': 'Float'
            },
            'placed': {
                'sdtype': 'boolean'
            },
            'ssn': {
                'sdtype': 'ssn',
                'pii': True
            },
            'credit_card': {
                'sdtype': 'credit_card_number',
                'pii': True
            },
            'drivers_license': {
                'sdtype': 'id',
                'regex_format': 'regex'
            }
        },
        'primary_key': 'student_id',
        'alternate_keys': ['drivers_license']
    }
    assert new_metadata == expected_metadata


@patch('sdv.metadata.metadata_upgrader._upgrade_constraints')
def test_convert_metadata_with_constraints(upgrade_constraints_mock):
    """Test the ``convert_metadata`` method with constraints.

    The method should take a dictionary of the old metadata format and convert it to the new
    format.

    Setup:
        - Mock the ``_upgrade_constraints`` method.

    Input:
        - Dictionary of single table metadata in the old schema and constraints.

    Output:
        - Dictionary of the same metadata with the new schema and constraints.
    """
    # Setup
    old_metadata = {
        'fields': {
            'salary': {
                'type': 'numerical',
                'subtype': 'integer'
            },
            'student_id': {
                'type': 'id',
                'subtype': 'integer'
            },
        },
        'primary_key': 'student_id',
        'constraints': [
            {
                'constraint': 'sdv.constraints.tabular.UniqueCombinations',
                'columns': ['a', 'b']
            },
            {
                'constraint': 'sdv.constraints.tabular.OneHotEncoding',
                'columns': ['c', 'd']
            }
        ]
    }
    new_constraints = [
        {
            'constraint_name': 'FixedCombinations',
            'column_names': ['a', 'b']
        },
        {
            'constraint_name': 'OneHotEncoding',
            'column_names': ['c', 'd']
        }
    ]
    upgrade_constraints_mock.return_value = new_constraints

    # Run
    new_metadata = convert_metadata(old_metadata)

    # Assert
    expected_metadata = {
        'columns': {
            'salary': {
                'sdtype': 'numerical',
                'computer_representation': 'Int64'
            },
            'student_id': {
                'sdtype': 'id',
                'regex_format': r'\d{30}'
            }
        },
        'primary_key': 'student_id'
    }
    assert new_metadata == expected_metadata



"""Formatter for numerical data."""
import logging
import sys

import numpy as np
import pandas as pd

LOGGER = logging.getLogger(__name__)

MAX_DECIMALS = sys.float_info.dig - 1
INTEGER_BOUNDS = {
    'Int8': (-2**7, 2**7 - 1),
    'Int16': (-2**15, 2**15 - 1),
    'Int32': (-2**31, 2**31 - 1),
    'Int64': (-2**63, 2**63 - 1),
    'UInt8': (0, 2**8 - 1),
    'UInt16': (0, 2**16 - 1),
    'UInt32': (0, 2**32 - 1),
    'UInt64': (0, 2**64 - 1),
}


class NumericalFormatter:
    """Formatter for numerical data.

    Args:
        enforce_rounding (bool):
            Whether or not to learn what place to round to based on the data seen during ``fit``.
            If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
            Defaults to ``False``.
        enforce_min_max_values (bool):
            Whether or not to clip the data returned by ``reverse_transform`` to the min and
            max values seen during ``fit``.
            Defaults to ``False``.
        computer_representation (dtype):
            Accepts ``'Int8'``, ``'Int16'``, ``'Int32'``, ``'Int64'``, ``'UInt8'``, ``'UInt16'``,
            ``'UInt32'``, ``'UInt64'``, ``'Float'``.
            Defaults to ``'Float'``.
    """

    _dtype = None
    _min_value = None
    _max_value = None
    _rounding_digits = None

    def __init__(self, enforce_rounding=False, enforce_min_max_values=False,
                 computer_representation='Float'):
        self.enforce_rounding = enforce_rounding
        self.enforce_min_max_values = enforce_min_max_values
        self.computer_representation = computer_representation

    @staticmethod
    def _learn_rounding_digits(data):
        """Check if data has any decimals."""
        name = data.name
        data = np.array(data)
        roundable_data = data[~(np.isinf(data) | pd.isna(data))]

        # Doesn't contain numbers
        if len(roundable_data) == 0:
            return None

        # Doesn't contain decimal digits
        if ((roundable_data % 1) == 0).all():
            return 0

        # Try to round to fewer digits
        if (roundable_data == roundable_data.round(MAX_DECIMALS)).all():
            for decimal in range(MAX_DECIMALS + 1):
                if (roundable_data == roundable_data.round(decimal)).all():
                    return decimal

        # Can't round, not equal after MAX_DECIMALS digits of precision
        LOGGER.info(
            f"No rounding scheme detected for column '{name}'."
            ' Synthetic data will not be rounded.'
        )
        return None

    def learn_format(self, column):
        """Learn the format of a column.

        Args:
            column (pandas.Series):
                Data to learn the format.
        """
        self._dtype = column.dtype
        if self.enforce_min_max_values:
            self._min_value = column.min()
            self._max_value = column.max()

        if self.enforce_rounding:
            self._rounding_digits = self._learn_rounding_digits(column)

    def format_data(self, column):
        """Format a column according to the learned format.

        Args:
            column (pd.Series):
                Data to format.

        Returns:
            numpy.ndarray:
                containing the formatted data.
        """
        column = column.copy().to_numpy()
        if self.enforce_min_max_values:
            column = column.clip(self._min_value, self._max_value)
        elif self.computer_representation != 'Float':
            min_bound, max_bound = INTEGER_BOUNDS[self.computer_representation]
            column = column.clip(min_bound, max_bound)

        is_integer = np.dtype(self._dtype).kind == 'i'
        if self.enforce_rounding and self._rounding_digits is not None:
            column = column.round(self._rounding_digits)
        elif is_integer:
            column = column.round(0)

        if pd.isna(column).any() and is_integer:
            return column

        return column.astype(self._dtype)



import pandas as pd

from sdv.constraints import create_custom_constraint_class


def is_valid(column_names, data):
    """Validate the constraint."""
    return pd.Series([value[0] > 1 for value in data[column_names].to_numpy()])


def transform(column_names, data):
    """Transform the constraint."""
    data[column_names] = data[column_names] ** 2
    return data


def reverse_transform(column_names, data):
    """Reverse transform the constraint."""
    data[column_names] = data[column_names] // 2
    return data


MyConstraint = create_custom_constraint_class(
    is_valid,
    transform,
    reverse_transform
)


def amenities_is_valid(column_names, data):
    """Validate that if ``has_rewards`` amenities fee is 0."""
    boolean_column = column_names[0]
    numerical_column = column_names[1]
    true_values = (data[boolean_column]) & (data[numerical_column] == 0.0)
    false_values = ~data[boolean_column]

    return (true_values) | (false_values)


def amenities_transform(column_names, data):
    """Transform the data if amenities fee is to be applied."""
    boolean_column = column_names[0]
    numerical_column = column_names[1]
    typical_value = data[numerical_column].median()
    data[numerical_column] = data[numerical_column].mask(
        data[boolean_column],
        typical_value
    )

    return data


def amenities_reverse_transform(column_names, data):
    """Reverse the data if amenities fee is to be applied."""
    boolean_column = column_names[0]
    numerical_column = column_names[1]
    data[numerical_column] = data[numerical_column].mask(data[boolean_column], 0.0)

    return data


IfTrueThenZero = create_custom_constraint_class(
    amenities_is_valid,
    amenities_transform,
    amenities_reverse_transform
)



"""Integration tests for the ``DataProcessor``."""
import itertools

import numpy as np
from rdt.transformers import (
    AnonymizedFaker, BinaryEncoder, FloatFormatter, LabelEncoder, RegexGenerator,
    UnixTimestampEncoder)

from sdv.data_processing import DataProcessor
from sdv.data_processing.datetime_formatter import DatetimeFormatter
from sdv.data_processing.numerical_formatter import NumericalFormatter
from sdv.datasets.demo import download_demo
from sdv.metadata import SingleTableMetadata
from sdv.utils import get_datetime_format


def test_data_processor_with_anonymized_columns():
    """Test the ``DataProcessor``.

    Test that when we update a ``pii`` column this uses ``AnonymizedFaker`` to create
    a new set of data for that column, while it drops it on the ``transform``.

    Setup:
        - Load metadata and data.
        - Parse the old metadata to a new metadata.
        - Anonymize the field ``occupation``.

    Run:
        - Create a ``DataProcessor`` with the new metadata.
        - Fit the ``DataProcessor`` instance.
        - Transform the data.
        - Reverse transform the data.

    Assert:
        - The column ``occupation`` has been dropped in the ``transform`` process.
        - The column ``occupation`` has been re-created with new values from the
          ``AnonymizedFaker`` instance.
    """
    # Load metadata and data
    data, metadata = download_demo('single_table', 'adult')

    # Add anonymized field
    metadata.update_column('occupation', sdtype='job', pii=True)

    # Instance ``DataProcessor``
    dp = DataProcessor(metadata)

    # Fit
    dp.fit(data)

    # Transform
    transformed = dp.transform(data)

    # Reverse Transform
    reverse_transformed = dp.reverse_transform(transformed)

    # Assert
    assert reverse_transformed.occupation.isin(data.occupation).sum() == 0
    assert 'occupation' not in transformed.columns


def test_data_processor_with_anonymized_columns_and_primary_key():
    """Test the ``DataProcessor``.

    Test that when we update a ``pii`` column this uses ``AnonymizedFaker`` to create
    a new set of data for that column, while it drops it on the ``transform`` and this values
    can be repeated. Meanwhile, when we set a primary key this has to have a unique length
    of the input data.

    Setup:
        - Load metadata and data.
        - Parse the old metadata to a new metadata.
        - Anonymize the field ``occupation``.
        - Create ``id``.

    Run:
        - Create a ``DataProcessor`` with the new metadata.
        - Fit the ``DataProcessor`` instance.
        - Transform the data.
        - Reverse transform the data.

    Assert:
        - The column ``occupation`` has been dropped in the ``transform`` process.
        - The column ``occupation`` has been re-created with new values from the
          ``AnonymizedFaker`` instance.
        - The column ``id`` has been created in ``transform`` with a unique length of the data.
    """
    # Load metadata and data
    data, metadata = download_demo('single_table', 'adult')

    # Add anonymized field
    metadata.update_column('occupation', sdtype='job', pii=True)

    # Add primary key field
    metadata.add_column('id', sdtype='id', regex_format='ID_\\d{4}[0-9]')
    metadata.set_primary_key('id')

    # Add id
    size = len(data)
    data['id'] = np.arange(0, size).astype('O')

    # Instance ``DataProcessor``
    dp = DataProcessor(metadata)

    # Fit
    dp.fit(data)

    # Transform
    transformed = dp.transform(data)

    # Reverse Transform
    reverse_transformed = dp.reverse_transform(transformed)

    # Assert
    assert transformed.index.name == 'id'
    assert reverse_transformed.occupation.isin(data.occupation).sum() == 0
    assert 'occupation' not in transformed.columns
    assert 'id' not in transformed.columns
    assert reverse_transformed.id.isin(data.id).sum() == 0
    assert len(reverse_transformed.id.unique()) == size


def test_data_processor_with_primary_key_numerical():
    """End to end test for the ``DataProcessor``.

    Test that when running the ``DataProcessor`` with a numerical primary key, this is able
    to generate such a key in the ``transform`` method. The key should be the same length
    as the input data size.

    Setup:
        - Load metadata and data.
        - Detect the metadata from the ``dataframe``.
        - Create ``id`` column.

    Run:
        - Create a ``DataProcessor`` with the new metadata.
        - Fit the ``DataProcessor`` instance.
        - Transform the data.
        - Reverse transform the data.

    Assert:
        - The column ``id`` has been created during ``transform`` with a unique length of the data.
          matching the original numerical ``id`` column.
    """
    # Load metadata and data
    data, _ = download_demo('single_table', 'adult')
    adult_metadata = SingleTableMetadata()
    adult_metadata.detect_from_dataframe(data=data)

    # Add primary key field
    adult_metadata.add_column('id', sdtype='id')
    adult_metadata.set_primary_key('id')

    # Add id
    size = len(data)
    id_generator = itertools.count()
    ids = [next(id_generator) for _ in range(size)]
    data['id'] = ids

    # Instance ``DataProcessor``
    dp = DataProcessor(adult_metadata)

    # Fit
    dp.fit(data)

    # Transform
    transformed = dp.transform(data)

    # Reverse Transform
    reverse_transformed = dp.reverse_transform(transformed)

    # Assert
    assert transformed.index.name == 'id'
    assert 'id' not in transformed.columns
    assert reverse_transformed.index.isin(data.id).sum() == size
    assert len(reverse_transformed.id.unique()) == size


def test_data_processor_with_alternate_keys():
    """Test that alternate keys are being generated in a unique way.

    Test that the alternate keys are being generated and dropped the same way
    as with the ``primary_key``.
    """
    # Load metadata and data
    data, _ = download_demo('single_table', 'adult')
    data['fnlwgt'] = data['fnlwgt'].astype(str)
    adult_metadata = SingleTableMetadata()
    adult_metadata.detect_from_dataframe(data=data)

    # Add primary key field
    adult_metadata.add_column('id', sdtype='id')
    adult_metadata.set_primary_key('id')

    adult_metadata.add_column('secondary_id', sdtype='id')
    adult_metadata.update_column('fnlwgt', sdtype='id', regex_format='ID_\\d{4}[0-9]')

    adult_metadata.add_alternate_keys(['secondary_id', 'fnlwgt'])

    # Add id
    size = len(data)
    id_generator = itertools.count()
    ids = [next(id_generator) for _ in range(size)]
    data['id'] = ids
    data['secondary_id'] = ids

    # Instance ``DataProcessor``
    dp = DataProcessor(adult_metadata)

    # Fit
    dp.fit(data)

    # Transform
    transformed = dp.transform(data)

    # Reverse Transform
    reverse_transformed = dp.reverse_transform(transformed)

    # Assert
    assert 'id' not in transformed.columns
    assert 'secondary_id' not in transformed.columns
    assert 'fnlwgt' not in transformed.columns
    assert len(reverse_transformed.id.unique()) == size
    assert len(reverse_transformed.secondary_id.unique()) == size
    assert len(reverse_transformed.fnlwgt.unique()) == size


def test_data_processor_prepare_for_fitting():
    """Test the ``prepare_for_fitting`` method.

    Test that the method sets an expected list of transformers for the given
    data types of the ``metadata`` and also respects the extra parameters
    that those have. In this case the columns ``start_date`` and ``end_date`` have
    a ``datetime_format`` which has to be set to the ``UnixTimestampEncoder`` and
    the column ``salary`` has a ``computer_representation`` set to ``Int64``.
    """
    # Setup
    data, metadata = download_demo(
        modality='single_table',
        dataset_name='student_placements_pii'
    )
    dp = DataProcessor(metadata)

    # Run
    dp.prepare_for_fitting(data)

    # Assert
    field_transformers = dp._hyper_transformer.field_transformers
    expected_transformers = {
        'mba_spec': LabelEncoder,
        'employability_perc': FloatFormatter,
        'placed': LabelEncoder,
        'student_id': RegexGenerator,
        'experience_years': FloatFormatter,
        'duration': LabelEncoder,
        'salary': FloatFormatter,
        'second_perc': FloatFormatter,
        'start_date': UnixTimestampEncoder,
        'address': AnonymizedFaker,
        'gender': LabelEncoder,
        'mba_perc': FloatFormatter,
        'degree_type': LabelEncoder,
        'end_date': UnixTimestampEncoder,
        'high_spec': LabelEncoder,
        'high_perc': FloatFormatter,
        'work_experience': LabelEncoder,
        'degree_perc': FloatFormatter
    }
    for column_name, transformer_class in expected_transformers.items():
        if transformer_class is not None:
            assert isinstance(field_transformers[column_name], transformer_class)
        else:
            assert field_transformers[column_name] is None

    assert field_transformers['start_date'].datetime_format == '%Y-%m-%d'
    assert field_transformers['end_date'].datetime_format == '%Y-%m-%d'
    assert field_transformers['salary'].computer_representation == 'Int64'


def test_data_processor_reverse_transform_with_formatters():
    """End to end test using formatters."""
    # Setup
    data, metadata = download_demo(
        modality='single_table',
        dataset_name='student_placements'
    )
    dp = DataProcessor(metadata)

    # Run
    dp.fit(data)

    transformed = dp.transform(data)
    reverse_transformed = dp.reverse_transform(transformed)
    reverse_transformed = reverse_transformed.drop('student_id', axis=1)
    reverse_transformed = reverse_transformed.reset_index()

    # Assert
    assert isinstance(dp.formatters['degree_perc'], NumericalFormatter)
    assert isinstance(dp.formatters['employability_perc'], NumericalFormatter)
    assert isinstance(dp.formatters['experience_years'], NumericalFormatter)
    assert isinstance(dp.formatters['high_perc'], NumericalFormatter)
    assert isinstance(dp.formatters['mba_perc'], NumericalFormatter)
    assert isinstance(dp.formatters['salary'], NumericalFormatter)
    assert isinstance(dp.formatters['second_perc'], NumericalFormatter)
    assert isinstance(dp.formatters['start_date'], DatetimeFormatter)
    assert isinstance(dp.formatters['end_date'], DatetimeFormatter)

    start_date_data_format = get_datetime_format(data['start_date'][~data['start_date'].isna()][0])
    reversed_start_date = reverse_transformed['start_date'][
        ~reverse_transformed['start_date'].isna()
    ]
    reversed_start_date_format = get_datetime_format(reversed_start_date.iloc[0])
    assert start_date_data_format == reversed_start_date_format

    end_date_data_format = get_datetime_format(data['end_date'][~data['end_date'].isna()][0])
    reversed_end_date = reverse_transformed['end_date'][~reverse_transformed['end_date'].isna()]
    reversed_end_date_format = get_datetime_format(reversed_end_date.iloc[0])
    assert end_date_data_format == reversed_end_date_format


def test_data_processor_refit_hypertransformer():
    """Test data processor re-fits _hyper_transformer."""
    # Setup
    data, metadata = download_demo(
        modality='single_table',
        dataset_name='student_placements'
    )
    dp = DataProcessor(metadata)

    # Run
    dp.fit(data)
    dp.update_transformers({'placed': BinaryEncoder()})

    # Assert
    assert dp._hyper_transformer._fitted
    assert dp._hyper_transformer._modified_config

    dp.fit(data)

    transformed = dp.transform(data)
    assert all(transformed.dtypes == float)



"""Utils for testing."""
import pandas as pd

from sdv.metadata.multi_table import MultiTableMetadata


class DataFrameMatcher:
    """Match a given Pandas DataFrame in a mock function call."""

    def __init__(self, df):
        self.df = df

    def __eq__(self, other):
        pd.testing.assert_frame_equal(self.df, other)
        return True


def get_multi_table_metadata():
    """Return a ``MultiTableMetadata`` object to be used with tests."""
    dict_metadata = {
        'tables': {
            'nesreca': {
                'primary_key': 'id_nesreca',
                'columns': {
                    'upravna_enota': {'sdtype': 'id'},
                    'id_nesreca': {'sdtype': 'id'},
                    'nesreca_val': {'sdtype': 'numerical'}
                }
            },
            'oseba': {
                'columns': {
                    'upravna_enota': {'sdtype': 'id'},
                    'id_nesreca': {'sdtype': 'id'},
                    'oseba_val': {'sdtype': 'numerical'}
                }
            },
            'upravna_enota': {
                'primary_key': 'id_upravna_enota',
                'columns': {
                    'id_upravna_enota': {'sdtype': 'id'},
                    'upravna_val': {'sdtype': 'numerical'}
                }
            }
        },
        'relationships': [
            {
                'parent_table_name': 'upravna_enota',
                'parent_primary_key': 'id_upravna_enota',
                'child_table_name': 'nesreca',
                'child_foreign_key': 'upravna_enota'
            },
            {
                'parent_table_name': 'upravna_enota',
                'parent_primary_key': 'id_upravna_enota',
                'child_table_name': 'oseba',
                'child_foreign_key': 'upravna_enota'
            },
            {
                'parent_table_name': 'nesreca',
                'parent_primary_key': 'id_nesreca',
                'child_table_name': 'oseba',
                'child_foreign_key': 'id_nesreca'
            }
        ],
        'METADATA_SPEC_VERSION': 'MULTI_TABLE_V1'
    }

    return MultiTableMetadata.load_from_dict(dict_metadata)


def get_multi_table_data():
    """Return a dictionary containing some data for multi table."""
    data = {
        'nesreca': pd.DataFrame({
            'id_nesreca': list(range(4)),
            'upravna_enota': list(range(4)),
            'nesreca_val': list(range(4))
        }),
        'oseba': pd.DataFrame({
            'upravna_enota': list(range(4)),
            'id_nesreca': list(range(4)),
            'oseba_val': list(range(4))
        }),
        'upravna_enota': pd.DataFrame({
            'id_upravna_enota': list(range(4)),
            'upravna_val': list(range(4))
        }),
    }

    return data



import datetime
from unittest.mock import patch

import pandas as pd
import pkg_resources
import pytest
from copulas.multivariate.gaussian import GaussianMultivariate
from rdt.transformers import AnonymizedFaker, FloatFormatter, LabelEncoder, RegexGenerator

from sdv.datasets.demo import download_demo
from sdv.metadata import SingleTableMetadata
from sdv.sampling import Condition
from sdv.single_table import (
    CopulaGANSynthesizer, CTGANSynthesizer, GaussianCopulaSynthesizer, TVAESynthesizer)
from sdv.single_table.base import BaseSingleTableSynthesizer
from tests.integration.single_table.custom_constraints import MyConstraint

METADATA = SingleTableMetadata.load_from_dict({
    'METADATA_SPEC_VERSION': 'SINGLE_TABLE_V1',
    'columns': {
        'column1': {
            'sdtype': 'numerical'
        },
        'column2': {
            'sdtype': 'numerical'
        },
        'column3': {
            'sdtype': 'numerical'
        }
    }
})

SYNTHESIZERS = [
    pytest.param(CTGANSynthesizer(METADATA, epochs=1, cuda=False), id='CTGANSynthesizer'),
    pytest.param(TVAESynthesizer(METADATA, epochs=1, cuda=False), id='TVAESynthesizer'),
    pytest.param(GaussianCopulaSynthesizer(METADATA), id='GaussianCopulaSynthesizer'),
    pytest.param(CopulaGANSynthesizer(METADATA, epochs=1, cuda=False), id='CopulaGANSynthesizer'),
]


def _isinstance_side_effect(*args, **kwargs):
    if isinstance(args[0], GaussianMultivariate):
        return True
    else:
        return isinstance(args[0], args[1])


@pytest.mark.parametrize('synthesizer', SYNTHESIZERS)
def test_conditional_sampling_graceful_reject_sampling_true_dict(synthesizer):
    data = pd.DataFrame({
        'column1': list(range(100)),
        'column2': list(range(100)),
        'column3': list(range(100))
    })

    synthesizer.fit(data)
    conditions = [
        Condition({
            'column1': 28,
            'column2': 37,
            'column3': 93
        })
    ]

    with pytest.raises(ValueError):  # noqa: PT011
        synthesizer.sample_from_conditions(conditions=conditions)


@pytest.mark.parametrize('synthesizer', SYNTHESIZERS)
def test_conditional_sampling_graceful_reject_sampling_true_dataframe(synthesizer):
    data = pd.DataFrame({
        'column1': list(range(100)),
        'column2': list(range(100)),
        'column3': list(range(100))
    })

    synthesizer.fit(data)
    conditions = pd.DataFrame({
        'column1': [28],
        'column2': [37],
        'column3': [93]
    })

    with pytest.raises(ValueError, match='a'):
        synthesizer.sample_remaining_columns(conditions)


def test_fit_with_unique_constraint_on_data_with_only_index_column():
    """Test that the ``fit`` method runs without error when metadata specifies unique constraint,
    ``fit`` is called on data containing a column named index.

    The ``fit`` method is expected to fit the model to data,
    taking into account the metadata and the ``Unique`` constraint.

    Setup:
    - The model is passed the unique constraint and
    the primary key column.

    Input:
    - Data, Unique constraint

    Github Issue:
    - Tests that https://github.com/sdv-dev/SDV/issues/616 does not occur
    """
    # Setup
    test_df = pd.DataFrame({
        'key': [
            1,
            2,
            3,
            4,
            5,
        ],
        'index': [
            'A',
            'B',
            'C',
            'D',
            'E',
        ]
    })

    metadata = SingleTableMetadata()
    metadata.add_column('key', sdtype='id')
    metadata.add_column('index', sdtype='categorical')
    metadata.set_primary_key('key')

    model = GaussianCopulaSynthesizer(metadata)
    constraint = {
        'constraint_class': 'Unique',
        'constraint_parameters': {
            'column_names': ['index']
        }
    }
    model.add_constraints([constraint])

    # Run
    model.fit(test_df)
    samples = model.sample(2)

    # Assert
    assert len(samples) == 2
    assert samples['index'].is_unique


def test_fit_with_unique_constraint_on_data_which_has_index_column():
    """Test that the ``fit`` method runs without error when metadata specifies unique constraint,
    ``fit`` is called on data containing a column named index and other columns.

    The ``fit`` method is expected to fit the model to data,
    taking into account the metadata and the ``Unique`` constraint.

    Setup:
    - The model is passed the unique constraint and
    the primary key column.
    - The unique constraint is set on the ``test_column``

    Input:
    - Data, Unique constraint

    Github Issue:
    - Tests that https://github.com/sdv-dev/SDV/issues/616 does not occur
    """
    # Setup
    test_df = pd.DataFrame({
        'key': [
            1,
            2,
            3,
            4,
            5,
        ],
        'index': [
            'A',
            'B',
            'C',
            'D',
            'E',
        ],
        'test_column': [
            'A1',
            'B2',
            'C3',
            'D4',
            'E5',
        ]
    })

    metadata = SingleTableMetadata()
    metadata.add_column('key', sdtype='id')
    metadata.add_column('index', sdtype='categorical')
    metadata.add_column('test_column', sdtype='categorical')
    metadata.set_primary_key('key')

    model = GaussianCopulaSynthesizer(metadata)
    constraint = {
        'constraint_class': 'Unique',
        'constraint_parameters': {
            'column_names': ['test_column']
        }
    }
    model.add_constraints([constraint])

    # Run
    model.fit(test_df)
    samples = model.sample(2)

    # Assert
    assert len(samples) == 2
    assert samples['test_column'].is_unique


def test_fit_with_unique_constraint_on_data_subset():
    """Test that the ``fit`` method runs without error when metadata specifies unique constraint,
    ``fit`` is called on a subset of the original data.

    The ``fit`` method is expected to fit the model to the subset of data,
    taking into account the metadata and the ``Unique`` constraint.

    Setup:
    - The model is passed a ``Unique`` constraint and is
    matched to a subset of the specified data.
    Subdividing the data results in missing indexes in the subset contained in the original data.

    Input:
    - Subset of data, unique constraint

    Github Issue:
    - Tests that https://github.com/sdv-dev/SDV/issues/610 does not occur
    """
    # Setup
    test_df = pd.DataFrame({
        'key': [
            1,
            2,
            3,
            4,
            5,
        ],
        'test_column': [
            'A',
            'B',
            'C',
            'D',
            'E',
        ]
    })

    metadata = SingleTableMetadata()
    metadata.add_column('key', sdtype='id')
    metadata.add_column('test_column', sdtype='categorical')
    metadata.set_primary_key('key')

    test_df = test_df.iloc[[1, 3, 4]]
    constraint = {
        'constraint_class': 'Unique',
        'constraint_parameters': {
            'column_names': ['test_column']
        }
    }
    model = GaussianCopulaSynthesizer(metadata)
    model.add_constraints([constraint])

    # Run
    model.fit(test_df)
    samples = model.sample(2)

    # Assert
    assert len(samples) == 2
    assert samples['test_column'].is_unique


@patch('sdv.single_table.base.isinstance')
@patch('sdv.single_table.copulas.multivariate.GaussianMultivariate',
       spec_set=GaussianMultivariate)
def test_conditional_sampling_constraint_uses_reject_sampling(gm_mock, isinstance_mock):
    """Test that the ``sample`` method handles constraints with conditions.

    The ``sample`` method is expected to properly apply constraint
    transformations by dropping columns that cannot be conditonally sampled
    on due to them being part of a constraint.

    Setup:
    - The model is being passed a ``UniqueCombination`` constraint and then
    asked to sample with two conditions, one of which the constraint depends on.
    The constraint is expected to skip its transformations since only some of
    the columns are provided by the conditions and the model will use reject
    sampling to meet the constraint instead.

    Input:
    - Conditions
    Side Effects:
    - Correct columns to condition on are passed to underlying sample method
    """
    # Setup
    isinstance_mock.side_effect = _isinstance_side_effect
    data = pd.DataFrame({
        'city': ['LA', 'SF', 'CHI', 'LA', 'LA'],
        'state': ['CA', 'CA', 'IL', 'CA', 'CA'],
        'age': [27, 28, 26, 21, 30]
    })

    metadata = SingleTableMetadata()
    metadata.add_column('city', sdtype='categorical')
    metadata.add_column('state', sdtype='categorical')
    metadata.add_column('age', sdtype='numerical')

    model = GaussianCopulaSynthesizer(metadata)

    constraint = {
        'constraint_class': 'FixedCombinations',
        'constraint_parameters': {
            'column_names': ['city', 'state']
        }
    }
    model.add_constraints([constraint])
    sampled_numeric_data = [
        pd.DataFrame({
            'city#state': [0, 1, 2, 0, 0],
            'age': [30, 30, 30, 30, 30]
        }),
        pd.DataFrame({
            'city#state': [1],
            'age': [30]
        })
    ]
    gm_mock.return_value.sample.side_effect = sampled_numeric_data
    model.fit(data)

    # Run
    conditions = [Condition({'age': 30, 'state': 'CA'}, num_rows=5)]
    sampled_data = model.sample_from_conditions(conditions=conditions)

    # Assert
    expected_transformed_conditions = {'age': 30}
    expected_data = pd.DataFrame({
        'city': ['LA', 'SF', 'LA', 'LA', 'SF'],
        'state': ['CA', 'CA', 'CA', 'CA', 'CA'],
        'age': [30, 30, 30, 30, 30]
    })
    sample_calls = model._model.sample.mock_calls
    assert len(sample_calls) == 2
    model._model.sample.assert_any_call(5, conditions=expected_transformed_conditions)
    pd.testing.assert_frame_equal(sampled_data, expected_data)


def test_sample_from_conditions_with_batch_size():
    """Test the ``sample_from_conditions`` method with a different ``batch_size``.

    If a smaller ``batch_size`` is passed, then the conditions should be broken down into
    batches of that size. If the ``batch_size`` is larger than the condition length, then
    the condition length should be used.

    - Input:
        - Conditions one of length 100 and another of length 10
        - Batch size of length 50

    - Output:
        - Sampled data
    """
    # Setup
    data = pd.DataFrame({
        'column1': list(range(100)),
        'column2': list(range(100)),
        'column3': list(range(100))
    })

    metadata = SingleTableMetadata()
    metadata.add_column('column1', sdtype='numerical')
    metadata.add_column('column2', sdtype='numerical')
    metadata.add_column('column3', sdtype='numerical')

    model = GaussianCopulaSynthesizer(metadata)
    model.fit(data)
    conditions = [
        Condition({'column1': 10}, num_rows=100),
        Condition({'column1': 50}, num_rows=10)
    ]

    # Run
    sampled_data = model.sample_from_conditions(conditions, batch_size=50)

    # Assert
    expected = pd.Series([10] * 100 + [50] * 10, name='column1')
    pd.testing.assert_series_equal(sampled_data['column1'], expected)


def test_multiple_fits():
    """Test the synthesizer refits correctly on new data.

    The synthesizer should refit the formatters and constraints.
    """
    # Setup
    data_1 = pd.DataFrame({
        'city': ['LA', 'SF', 'CHI', 'LA', 'LA'],
        'state': ['CA', 'CA', 'IL', 'CA', 'CA'],
        'measurement': [27.123, 28.756, 26.908, 21.002, 30.987]
    })
    data_2 = pd.DataFrame({
        'city': ['LA', 'LA', 'CHI', 'LA', 'LA'],
        'state': ['CA', 'CA', 'IL', 'CA', 'CA'],
        'measurement': [27.1, 28.7, 26.9, 21.2, 30.9]
    })
    metadata = SingleTableMetadata()
    metadata.add_column('city', sdtype='categorical')
    metadata.add_column('state', sdtype='categorical')
    metadata.add_column('measurement', sdtype='numerical')
    constraint = {
        'constraint_class': 'FixedCombinations',
        'constraint_parameters': {
            'column_names': ['city', 'state']
        }
    }
    model = GaussianCopulaSynthesizer(metadata)
    model.add_constraints([constraint])

    # Run
    model.fit(data_1)
    model.fit(data_2)

    # Assert
    assert ('SF', 'CA') not in model._data_processor._constraints[0]._combinations_to_uuids
    assert model._data_processor.formatters['measurement']._rounding_digits == 1


@pytest.mark.parametrize('synthesizer', SYNTHESIZERS)
def test_sampling(synthesizer):
    """Test that samples are different when ``reset_sampling`` is not called."""
    sample_1 = synthesizer.sample(10)
    sample_2 = synthesizer.sample(10)

    with pytest.raises(AssertionError):
        pd.testing.assert_frame_equal(sample_1, sample_2)


@pytest.mark.parametrize('synthesizer', SYNTHESIZERS)
def test_sampling_reset_sampling(synthesizer):
    """Test ``sample`` method for each synthesizer using ``reset_sampling``."""
    metadata = SingleTableMetadata.load_from_dict({
        'METADATA_SPEC_VERSION': 'SINGLE_TABLE_V1',
        'columns': {
            'column1': {
                'sdtype': 'numerical'
            },
            'column2': {
                'sdtype': 'address'
            },
            'column3': {
                'sdtype': 'email'
            },
            'column4': {
                'sdtype': 'ssn',
                'pii': True
            }
        }
    })
    data = pd.DataFrame({
        'column1': list(range(100)),
        'column2': [str(i) for i in (range(100))],
        'column3': [str(i) for i in (range(100))],
        'column4': [str(i) for i in (range(100))],
    })

    if isinstance(synthesizer, (CTGANSynthesizer, TVAESynthesizer)):
        synthesizer = synthesizer.__class__(metadata, cuda=False)
    else:
        synthesizer = synthesizer.__class__(metadata)

    synthesizer.fit(data)

    sampled1 = synthesizer.sample(10)
    synthesizer.reset_sampling()
    sampled2 = synthesizer.sample(10)

    pd.testing.assert_frame_equal(sampled1, sampled2)


def test_config_creation_doesnt_raise_error():
    """Test https://github.com/sdv-dev/SDV/issues/1110."""
    # Setup
    test_data = pd.DataFrame({
        'address_col': ['223 Williams Rd', '75 Waltham St', '77 Mass Ave'],
        'numerical_col': [1, 2, 3],
    })
    test_metadata = SingleTableMetadata()

    # Run
    test_metadata.detect_from_dataframe(test_data)
    test_metadata.update_column(
        column_name='address_col',
        sdtype='address',
        pii=False
    )

    synthesizer = GaussianCopulaSynthesizer(test_metadata)
    synthesizer.fit(test_data)


def test_transformers_correctly_auto_assigned():
    """Ensure the correct transformers and parameters are auto assigned to the data."""
    # Setup
    data = pd.DataFrame({
        'primary_key': ['user-000', 'user-001', 'user-002'],
        'pii_col': ['223 Williams Rd', '75 Waltham St', '77 Mass Ave'],
        'numerical_col': [1, 2, 3],
        'categorical_col': ['a', 'b', 'a'],
    })

    metadata = SingleTableMetadata()
    metadata.detect_from_dataframe(data)
    metadata.update_column(column_name='primary_key', sdtype='id', regex_format='user-[0-9]{3}')
    metadata.set_primary_key('primary_key')
    metadata.update_column(column_name='pii_col', sdtype='address', pii=True)
    synthesizer = GaussianCopulaSynthesizer(
        metadata, enforce_min_max_values=False, enforce_rounding=False)

    # Run
    synthesizer.auto_assign_transformers(data)
    transformers = synthesizer.get_transformers()

    # Assert
    assert isinstance(transformers['numerical_col'], FloatFormatter)
    assert isinstance(transformers['pii_col'], AnonymizedFaker)
    assert isinstance(transformers['primary_key'], RegexGenerator)
    assert isinstance(transformers['categorical_col'], LabelEncoder)

    assert transformers['numerical_col'].missing_value_replacement == 'mean'
    assert transformers['numerical_col'].model_missing_values is False
    assert transformers['numerical_col'].learn_rounding_scheme is False
    assert transformers['numerical_col'].enforce_min_max_values is False

    assert transformers['pii_col'].provider_name == 'address'
    assert transformers['pii_col'].function_name == 'address'

    assert transformers['primary_key'].regex_format == 'user-[0-9]{3}'
    assert transformers['primary_key'].enforce_uniqueness is True

    assert transformers['categorical_col'].add_noise is True


def test_custom_constraints_from_file(tmpdir):
    """Ensure the correct loading for a custom constraint class defined in another file."""
    data = pd.DataFrame({
        'primary_key': ['user-000', 'user-001', 'user-002'],
        'pii_col': ['223 Williams Rd', '75 Waltham St', '77 Mass Ave'],
        'numerical_col': [2, 3, 4],
        'categorical_col': ['a', 'b', 'a'],
    })

    metadata = SingleTableMetadata()
    metadata.detect_from_dataframe(data)
    metadata.update_column(column_name='pii_col', sdtype='address', pii=True)
    synthesizer = GaussianCopulaSynthesizer(
        metadata,
        enforce_min_max_values=False,
        enforce_rounding=False
    )
    synthesizer.load_custom_constraint_classes(
        'tests/integration/single_table/custom_constraints.py',
        ['MyConstraint']
    )
    constraint = {
        'constraint_class': 'MyConstraint',
        'constraint_parameters': {
            'column_names': ['numerical_col']
        }
    }

    # Run
    synthesizer.add_constraints([constraint])
    processed_data = synthesizer.preprocess(data)

    # Assert Processed Data
    assert all(processed_data['numerical_col'] == data['numerical_col'] ** 2)

    # Run - Fit the model
    synthesizer.fit_processed_data(processed_data)

    # Run - sample
    sampled = synthesizer.sample(10)
    assert all(sampled['numerical_col'] > 1)

    # Run - Save and Sample
    synthesizer.save(tmpdir / 'test.pkl')
    loaded_instance = synthesizer.load(tmpdir / 'test.pkl')
    loaded_sampled = loaded_instance.sample(10)
    assert all(loaded_sampled['numerical_col'] > 1)


def test_custom_constraints_from_object(tmpdir):
    """Ensure the correct loading for a custom constraint class passed as an object."""
    data = pd.DataFrame({
        'primary_key': ['user-000', 'user-001', 'user-002'],
        'pii_col': ['223 Williams Rd', '75 Waltham St', '77 Mass Ave'],
        'numerical_col': [2, 3, 4],
        'categorical_col': ['a', 'b', 'a'],
    })

    metadata = SingleTableMetadata()
    metadata.detect_from_dataframe(data)
    metadata.update_column(column_name='pii_col', sdtype='address', pii=True)
    synthesizer = GaussianCopulaSynthesizer(
        metadata,
        enforce_min_max_values=False,
        enforce_rounding=False
    )
    synthesizer.add_custom_constraint_class(MyConstraint, 'MyConstraint')
    constraint = {
        'constraint_class': 'MyConstraint',
        'constraint_parameters': {
            'column_names': ['numerical_col']
        }
    }

    # Run
    synthesizer.add_constraints([constraint])
    processed_data = synthesizer.preprocess(data)

    # Assert Processed Data
    assert all(processed_data['numerical_col'] == data['numerical_col'] ** 2)

    # Run - Fit the model
    synthesizer.fit_processed_data(processed_data)

    # Run - sample
    sampled = synthesizer.sample(10)
    assert all(sampled['numerical_col'] > 1)

    # Run - Save and Sample
    synthesizer.save(tmpdir / 'test.pkl')
    loaded_instance = synthesizer.load(tmpdir / 'test.pkl')
    loaded_sampled = loaded_instance.sample(10)
    assert all(loaded_sampled['numerical_col'] > 1)


def test_auto_assign_transformers_and_update_with_pii():
    """Ensure the ability to update a transformer with any given ``pii`` sdtype.

    This test is designed to auto-assign the transformers to an updated metadata that contains
    an ``pii`` field set as ``sdtype`` but no ``pii`` in the ``metadata`` itself. This should
    still assign the expected transformer to it.
    """
    # Setup
    data = pd.DataFrame(data={
        'id': ['N', 'A', 'K', 'F', 'P'],
        'numerical': [1, 2, 3, 2, 1],
        'name': ['A', 'A', 'B', 'B', 'B']
    })

    metadata = SingleTableMetadata()
    metadata.detect_from_dataframe(data)

    # Run
    metadata.update_column(column_name='id', sdtype='first_name')
    metadata.update_column(column_name='name', sdtype='name')
    metadata.set_primary_key('id')
    synthesizer = GaussianCopulaSynthesizer(metadata)
    synthesizer.auto_assign_transformers(data)

    # Assert
    id_transformer = synthesizer.get_transformers()['id']
    name_transformer = synthesizer.get_transformers()['name']
    assert id_transformer.provider_name == 'person'
    assert id_transformer.function_name == 'first_name'
    assert id_transformer.enforce_uniqueness is True

    assert name_transformer.provider_name == 'person'
    assert name_transformer.function_name == 'name'
    assert name_transformer.enforce_uniqueness is False


def test_refitting_a_model():
    """Test that refitting a model resets the sampling state of the generators."""
    # Setup
    data = pd.DataFrame(data={
        'id': [0, 1, 2, 3, 4],
        'numerical': [1, 2, 3, 2, 1],
        'name': ['A', 'A', 'B', 'B', 'B']
    })

    metadata = SingleTableMetadata()
    metadata.detect_from_dataframe(data)
    metadata.update_column(column_name='name', sdtype='name')
    metadata.update_column('id', sdtype='id')
    metadata.set_primary_key('id')

    synthesizer = GaussianCopulaSynthesizer(metadata)
    synthesizer.fit(data)
    first_sample = synthesizer.sample(10)

    # Run
    synthesizer.fit(data)
    second_sample = synthesizer.sample(10)

    # Assert
    assert all(second_sample['name'] == first_sample['name'])
    assert all(second_sample['id'] == first_sample['id'])


def test_get_info():
    """Test the correct dictionary is returned.

    Check the return dictionary is valid both before and after fitting a synthesizer.
    """
    # Setup
    data = pd.DataFrame({'col': [1, 2, 3]})
    today = datetime.datetime.today().strftime('%Y-%m-%d')
    metadata = SingleTableMetadata()
    metadata.add_column('col', sdtype='numerical')
    synthesizer = GaussianCopulaSynthesizer(metadata)

    # Run
    info = synthesizer.get_info()

    # Assert
    assert info == {
        'class_name': 'GaussianCopulaSynthesizer',
        'creation_date': today,
        'is_fit': False,
        'last_fit_date': None,
        'fitted_sdv_version': None
    }

    # Run
    synthesizer.fit(data)
    info = synthesizer.get_info()

    # Assert
    version = pkg_resources.get_distribution('sdv').version
    assert info == {
        'class_name': 'GaussianCopulaSynthesizer',
        'creation_date': today,
        'is_fit': True,
        'last_fit_date': today,
        'fitted_sdv_version': version
    }


def test_synthesizer_with_inequality_constraint():
    """Ensure that the ``Inequality`` constraint can sample from the model."""
    # Setup
    real_data, metadata = download_demo(
        modality='single_table',
        dataset_name='fake_hotel_guests'
    )
    synthesizer = GaussianCopulaSynthesizer(metadata)
    checkin_lessthan_checkout = {
        'constraint_class': 'Inequality',
        'constraint_parameters': {
            'low_column_name': 'checkin_date',
            'high_column_name': 'checkout_date'
        }
    }

    synthesizer.add_constraints([checkin_lessthan_checkout])
    synthesizer.fit(real_data)

    # Run and Assert
    sampled = synthesizer.sample(num_rows=500)
    synthesizer.validate(sampled)
    _sampled = sampled[~sampled['checkout_date'].isna()]
    assert all(
        pd.to_datetime(_sampled['checkin_date']) < pd.to_datetime(_sampled['checkout_date'])
    )


def test_save_and_load(tmp_path):
    """Test that synthesizers can be saved and loaded properly."""
    # Setup
    metadata = SingleTableMetadata()
    instance = BaseSingleTableSynthesizer(metadata)
    synthesizer_path = tmp_path / 'synthesizer.pkl'
    instance.save(synthesizer_path)

    # Run
    loaded_instance = BaseSingleTableSynthesizer.load(synthesizer_path)

    # Assert
    assert isinstance(loaded_instance, BaseSingleTableSynthesizer)
    assert instance.metadata.columns == {}
    assert instance.metadata.primary_key is None
    assert instance.metadata.alternate_keys == []
    assert instance.metadata.sequence_key is None
    assert instance.metadata.sequence_index is None
    assert instance.metadata._version == 'SINGLE_TABLE_V1'



"""Constraint utility functions."""

from datetime import datetime
from decimal import Decimal

import numpy as np
import pandas as pd


def cast_to_datetime64(value):
    """Cast a given value to a ``numpy.datetime64`` format.

    Args:
        value (pandas.Series, np.ndarray, list, or str):
            Input data to convert to ``numpy.datetime64``.

    Return:
        ``numpy.datetime64`` value or values.
    """
    if isinstance(value, str):
        value = pd.to_datetime(value).to_datetime64()
    elif isinstance(value, pd.Series):
        value.apply(lambda x: pd.to_datetime(x).to_datetime64())
        value = value.astype('datetime64[ns]')
    elif isinstance(value, (np.ndarray, list)):
        value = np.array([
            pd.to_datetime(item).to_datetime64()
            for item in value
        ])

    return value


def matches_datetime_format(value, datetime_format):
    """Check if datetime value matches the provided format.

    Args:
        value (str):
            The datetime value.
        datetime_format (str):
            The datetime format to check for.

    Return:
        True if the value matches the format. Otherwise False.
    """
    try:
        datetime.strptime(value, datetime_format)
    except Exception:
        return False

    return True


def _cast_to_type(data, dtype):
    if isinstance(data, pd.Series):
        data = data.apply(dtype)
    elif isinstance(data, (np.ndarray, list)):
        data = np.array([dtype(value) for value in data])
    else:
        data = dtype(data)

    return data


def logit(data, low, high):
    """Apply a logit function to the data using ``low`` and ``high``.

    Args:
        data (pd.Series, pd.DataFrame, np.array, int, float or datetime):
            Data to apply the logit function to.
        low (pd.Series, np.array, int, float or datetime):
            Low value/s to use when scaling.
        high (pd.Series, np.array, int, float or datetime):
            High value/s to use when scaling.

    Returns:
        Logit scaled version of the input data.
    """
    data = (data - low) / (high - low)
    data = _cast_to_type(data, Decimal)
    data = data * Decimal(0.95) + Decimal(0.025)
    data = _cast_to_type(data, float)
    return np.log(data / (1.0 - data))


def sigmoid(data, low, high):
    """Apply a sigmoid function to the data using ``low`` and ``high``.

    Args:
        data (pd.Series, pd.DataFrame, np.array, int, float or datetime):
            Data to apply the logit function to.
        low (pd.Series, np.array, int, float or datetime):
            Low value/s to use when scaling.
        high (pd.Series, np.array, int, float or datetime):
            High value/s to use when scaling.

    Returns:
        Sigmoid transform of the input data.
    """
    data = 1 / (1 + np.exp(-data))
    data = _cast_to_type(data, Decimal)
    data = (data - Decimal(0.025)) / Decimal(0.95)
    data = _cast_to_type(data, float)
    data = data * (high - low) + low

    return data




import pandas as pd

from sdv.evaluation.single_table import evaluate_quality, run_diagnostic
from sdv.metadata.single_table import SingleTableMetadata
from sdv.single_table.copulas import GaussianCopulaSynthesizer


def test_evaluation():
    """Test ``evaluate_quality`` and ``run_diagnostic``."""
    # Setup
    data = pd.DataFrame({'col': [1, 2, 3]})
    metadata = SingleTableMetadata()
    metadata.add_column('col', sdtype='numerical')
    synthesizer = GaussianCopulaSynthesizer(metadata)

    # Run and Assert
    synthesizer.fit(data)
    samples = synthesizer.sample(10)
    score = evaluate_quality(data, samples, metadata).get_score()
    assert score == 0.8333333333333334

    diagnostic = run_diagnostic(data, samples, metadata).get_results()
    assert diagnostic == {
        'DANGER': ['More than 50% of the synthetic rows are copies of the real data'],
        'SUCCESS': [
            'The synthetic data covers over 90% of the numerical ranges present in the real data',
            'The synthetic data follows over 90% of the min/max boundaries set by the real data'
        ],
        'WARNING': []
    }
